from typing import Optional

import ujson
from prefect import Flow

from infrahub.message_bus import RPCErrorResponse, messages
from infrahub.message_bus.operations import (
    check,
    event,
    finalize,
    git,
    refresh,
    requests,
    send,
)
from infrahub.message_bus.types import MessageTTL
from infrahub.services import InfrahubServices
from infrahub.tasks.check import set_check_status

COMMAND_MAP = {
    "check.artifact.create": check.artifact.create,
    "check.generator.run": check.generator.run,
    "check.repository.check_definition": check.repository.check_definition,
    "check.repository.merge_conflicts": check.repository.merge_conflicts,
    "check.repository.user_check": check.repository.user_check,
    "event.branch.merge": event.branch.merge,
    "event.node.mutated": event.node.mutated,
    "event.schema.update": event.schema.update,
    "event.worker.new_primary_api": event.worker.new_primary_api,
    "finalize.validator.execution": finalize.validator.execution,
    "git.file.get": git.file.get,
    "git.repository.connectivity": git.repository.connectivity,
    "refresh.git.fetch": git.repository.fetch,
    "refresh.registry.branches": refresh.registry.branches,
    "refresh.registry.rebased_branch": refresh.registry.rebased_branch,
    "request.generator_definition.check": requests.generator_definition.check,
    "request.artifact_definition.check": requests.artifact_definition.check,
    "request.proposed_change.pipeline": requests.proposed_change.pipeline,
    "request.proposed_change.refresh_artifacts": requests.proposed_change.refresh_artifacts,
    "request.repository.checks": requests.repository.checks,
    "request.repository.user_checks": requests.repository.user_checks,
    "send.echo.request": send.echo.request,
}


async def execute_message(
    routing_key: str, message_body: bytes, service: InfrahubServices, skip_flow: bool = False
) -> Optional[MessageTTL]:
    message_data = ujson.loads(message_body)
    message = messages.MESSAGE_MAP[routing_key](**message_data)
    message.set_log_data(routing_key=routing_key)
    try:
        func = COMMAND_MAP[routing_key]
        if skip_flow and isinstance(func, Flow):
            func = func.fn
        await func(message=message, service=service)
    except Exception as exc:  # pylint: disable=broad-except
        if message.reply_requested:
            response = RPCErrorResponse(errors=[str(exc)], initial_message=message.model_dump())
            await service.reply(message=response, initiator=message)
            return None
        if message.reached_max_retries:
            service.log.exception("Message failed after maximum number of retries", error=exc)
            await set_check_status(message, conclusion="failure", service=service)
            return None
        message.increase_retry_count()
        await service.send(message, delay=MessageTTL.FIVE, is_retry=True)
        return MessageTTL.FIVE



from __future__ import annotations

import ssl
from functools import cached_property
from typing import TYPE_CHECKING, Any

import httpx

from infrahub import config
from infrahub.exceptions import HTTPServerError, HTTPServerSSLError, HTTPServerTimeoutError
from infrahub.services.adapters.http import InfrahubHTTP

if TYPE_CHECKING:
    from infrahub.services import InfrahubServices


class HttpxAdapter(InfrahubHTTP):
    settings: config.HTTPSettings
    service: InfrahubServices

    async def initialize(self, service: InfrahubServices) -> None:
        """Initialize the HTTP adapter"""
        self.service = service
        self.settings = config.SETTINGS.http

        # Cache the context during init, this is to avoid issue when a CA bundle might be accessible
        # when Infrahub initializes but then removed before the first external HTTP call is made.
        _ = self.tls_context

    @cached_property
    def tls_context(self) -> ssl.SSLContext:
        return self.settings.get_tls_context()

    def verify_tls(self, verify: bool | None = None) -> bool | ssl.SSLContext:
        if verify is False:
            return False

        return self.tls_context

    async def _request(
        self,
        method: str,
        url: str,
        data: Any | None = None,
        json: Any | None = None,
        headers: dict[str, Any] | None = None,
        verify: bool | None = None,
    ) -> httpx.Response:
        """Returns an httpx.Response object or raises HTTPServerError or child classes."""
        params: dict[str, Any] = {}
        if data:
            params["data"] = data
        if json:
            params["json"] = json
        async with httpx.AsyncClient(verify=self.verify_tls(verify=verify)) as client:
            try:
                response = await client.request(
                    method=method,
                    url=url,
                    headers=headers,
                    timeout=self.settings.timeout,
                    **params,
                )
            except ssl.SSLCertVerificationError as exc:
                self.service.log.info(f"TLS verification failed for connection to {url}")
                raise HTTPServerSSLError(message=f"Unable to validate TLS certificate for connection to {url}") from exc
            except httpx.ReadTimeout as exc:
                self.service.log.info(f"Connection timed out when trying to reach {url}")
                raise HTTPServerTimeoutError(
                    message=f"Connection to {url} timed out after {self.settings.timeout}"
                ) from exc
            except httpx.RequestError as exc:
                # Catch all error from httpx
                self.service.log.warning(f"Unhandled HTTP error for {url} ({exc})")
                raise HTTPServerError(message=f"Unknown http error when connecting to {url}") from exc

        return response

    async def get(
        self,
        url: str,
        headers: dict[str, Any] | None = None,
    ) -> httpx.Response:
        return await self._request(
            method="get",
            url=url,
            headers=headers,
        )

    async def post(
        self,
        url: str,
        data: Any | None = None,
        json: Any | None = None,
        headers: dict[str, Any] | None = None,
        verify: bool | None = None,
    ) -> httpx.Response:
        return await self._request(method="post", url=url, data=data, json=json, headers=headers, verify=verify)



def get_commit_update_mutation(is_read_only: bool = False) -> str:
    mutation_commit_update_base = """
    mutation ($repository_id: String!, $commit: String!) {{
        {repo_class}Update(data: {{ id: $repository_id, commit: {{ is_protected: true, source: $repository_id, value: $commit }} }}) {{
            ok
            object {{
                commit {{
                    value
                }}
            }}
        }}
    }}
    """
    if is_read_only:
        return mutation_commit_update_base.format(repo_class="CoreReadOnlyRepository")
    return mutation_commit_update_base.format(repo_class="CoreRepository")


QUERY_RELATIONSHIPS = """
    query GetRelationships($relationship_identifiers: [String!]!) {
        Relationship(ids: $relationship_identifiers) {
            count
            edges {
                node {
                    identifier
                    peers {
                        id
                        kind
                    }
                }
            }
        }
    }
"""

SCHEMA_HASH_SYNC_STATUS = """
query {
  InfrahubStatus {
    summary {
      schema_hash_synced
    }
  }
}
"""

QUERY_USER = """
query GET_PROFILE_DETAILS {
  AccountProfile {
    id
    display_label
    account_type {
      value
      __typename
      updated_at
    }
    status {
      label
      value
      updated_at
      __typename
    }
    description {
      value
      updated_at
      __typename
    }
    label {
      value
      updated_at
      __typename
    }
    member_of_groups {
      count
      edges {
        node {
          display_label
          group_type {
            value
          }
          ... on CoreAccountGroup {
            id
            roles {
              count
              edges {
                node {
                  permissions {
                    count
                    edges {
                      node {
                        display_label
                        identifier {
                          value
                        }
                      }
                    }
                  }
                }
              }
            }
            display_label
          }
        }
      }
    }
    __typename
    name {
      value
      updated_at
      __typename
    }
  }
}
"""



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional

from infrahub.core import registry
from infrahub.core.constants import NULL_VALUE
from infrahub.core.schema import (
    MainSchemaTypes,
    SchemaAttributePath,
    SchemaAttributePathValue,
)
from infrahub.core.validators.uniqueness.index import UniquenessQueryResultsIndex
from infrahub.core.validators.uniqueness.model import (
    NodeUniquenessQueryRequest,
    QueryAttributePath,
    QueryRelationshipAttributePath,
)
from infrahub.core.validators.uniqueness.query import NodeUniqueAttributeConstraintQuery
from infrahub.exceptions import ValidationError

from .interface import NodeConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.query import QueryResult
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class NodeGroupedUniquenessConstraint(NodeConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch
        self.schema_branch = registry.schema.get_schema_branch(branch.name)

    async def _build_query_request(
        self,
        updated_node: Node,
        node_schema: MainSchemaTypes,
        path_groups: list[list[SchemaAttributePath]],
        filters: Optional[list[str]] = None,
    ) -> NodeUniquenessQueryRequest:
        query_request = NodeUniquenessQueryRequest(kind=node_schema.kind)
        for path_group in path_groups:
            include_in_query = not filters
            query_relationship_paths: set[QueryRelationshipAttributePath] = set()
            query_attribute_paths: set[QueryAttributePath] = set()
            for attribute_path in path_group:
                if attribute_path.related_schema and attribute_path.relationship_schema:
                    if filters and attribute_path.relationship_schema.name in filters:
                        include_in_query = True

                    relationship_manager: RelationshipManager = getattr(
                        updated_node, attribute_path.relationship_schema.name
                    )
                    related_node = await relationship_manager.get_peer(db=self.db)
                    related_node_id = related_node.get_id() if related_node else None
                    query_relationship_paths.add(
                        QueryRelationshipAttributePath(
                            identifier=attribute_path.relationship_schema.get_identifier(),
                            value=related_node_id,
                        )
                    )
                    continue
                if attribute_path.attribute_schema:
                    if filters and attribute_path.attribute_schema.name in filters:
                        include_in_query = True
                    attribute_name = attribute_path.attribute_schema.name
                    attribute = getattr(updated_node, attribute_name)
                    if attribute.is_enum and attribute.value:
                        attribute_value = attribute.value.value
                    else:
                        attribute_value = attribute.value
                    if attribute_value is None:
                        attribute_value = NULL_VALUE
                    query_attribute_paths.add(
                        QueryAttributePath(
                            attribute_name=attribute_name,
                            property_name=attribute_path.attribute_property_name or "value",
                            value=attribute_value,
                        )
                    )
            if include_in_query:
                query_request.relationship_attribute_paths |= query_relationship_paths
                query_request.unique_attribute_paths |= query_attribute_paths
        return query_request

    async def _get_node_attribute_path_values(
        self,
        updated_node: Node,
        path_group: list[SchemaAttributePath],
    ) -> list[SchemaAttributePathValue]:
        node_value_combination = []
        for schema_attribute_path in path_group:
            if schema_attribute_path.relationship_schema:
                relationship_name = schema_attribute_path.relationship_schema.name
                relationship_manager: RelationshipManager = getattr(updated_node, relationship_name)
                related_node = await relationship_manager.get_peer(db=self.db)
                related_node_id = related_node.get_id() if related_node else None
                node_value_combination.append(
                    SchemaAttributePathValue.from_schema_attribute_path(schema_attribute_path, value=related_node_id)
                )
            elif schema_attribute_path.attribute_schema:
                attribute_name = schema_attribute_path.attribute_schema.name
                attribute_field = getattr(updated_node, attribute_name)
                attribute_value = getattr(attribute_field, schema_attribute_path.attribute_property_name or "value")
                if attribute_field.is_enum and attribute_value:
                    attribute_value = attribute_value.value
                elif attribute_value is None:
                    attribute_value = NULL_VALUE
                node_value_combination.append(
                    SchemaAttributePathValue.from_schema_attribute_path(
                        schema_attribute_path,
                        value=attribute_value,
                    )
                )
        return node_value_combination

    def _check_one_constraint_group(
        self, schema_attribute_path_values: list[SchemaAttributePathValue], results_index: UniquenessQueryResultsIndex
    ) -> None:
        # constraint cannot be violated if this node is missing any values
        if any(sapv.value is None for sapv in schema_attribute_path_values):
            return

        matching_node_ids = results_index.get_node_ids_for_value_group(schema_attribute_path_values)
        if not matching_node_ids:
            return
        uniqueness_constraint_fields = []
        for sapv in schema_attribute_path_values:
            if sapv.relationship_schema:
                uniqueness_constraint_fields.append(sapv.relationship_schema.name)
            elif sapv.attribute_schema:
                uniqueness_constraint_fields.append(sapv.attribute_schema.name)
        uniqueness_constraint_string = "-".join(uniqueness_constraint_fields)
        error_msg = f"Violates uniqueness constraint '{uniqueness_constraint_string}'"
        errors = [ValidationError({field_name: error_msg}) for field_name in uniqueness_constraint_fields]
        raise ValidationError(errors)

    async def _check_results(
        self,
        updated_node: Node,
        path_groups: list[list[SchemaAttributePath]],
        query_results: Iterable[QueryResult],
    ) -> None:
        results_index = UniquenessQueryResultsIndex(
            query_results=query_results, exclude_node_ids={updated_node.get_id()}
        )
        for path_group in path_groups:
            schema_attribute_path_values = await self._get_node_attribute_path_values(
                updated_node=updated_node, path_group=path_group
            )
            self._check_one_constraint_group(
                schema_attribute_path_values=schema_attribute_path_values, results_index=results_index
            )

    async def _check_one_schema(
        self,
        node: Node,
        node_schema: MainSchemaTypes,
        at: Optional[Timestamp] = None,
        filters: Optional[list[str]] = None,
    ) -> None:
        schema_branch = self.db.schema.get_schema_branch(name=self.branch.name)
        path_groups = node_schema.get_unique_constraint_schema_attribute_paths(schema_branch=schema_branch)
        query_request = await self._build_query_request(
            updated_node=node, node_schema=node_schema, path_groups=path_groups, filters=filters
        )
        if not query_request:
            return
        query = await NodeUniqueAttributeConstraintQuery.init(
            db=self.db, branch=self.branch, at=at, query_request=query_request, min_count_required=0
        )
        await query.execute(db=self.db)
        await self._check_results(updated_node=node, path_groups=path_groups, query_results=query.get_results())

    async def check(self, node: Node, at: Optional[Timestamp] = None, filters: Optional[list[str]] = None) -> None:
        def _frozen_constraints(schema: MainSchemaTypes) -> frozenset[frozenset[str]]:
            if not schema.uniqueness_constraints:
                return frozenset()
            return frozenset(frozenset(uc) for uc in schema.uniqueness_constraints)

        node_schema = node.get_schema()
        include_node_schema = True
        frozen_node_constraints = _frozen_constraints(node_schema)
        schemas_to_check: list[MainSchemaTypes] = []
        if node_schema.inherit_from:
            for parent_schema_name in node_schema.inherit_from:
                parent_schema = self.schema_branch.get(name=parent_schema_name, duplicate=False)
                if not parent_schema.uniqueness_constraints:
                    continue
                schemas_to_check.append(parent_schema)
                frozen_parent_constraints = _frozen_constraints(parent_schema)
                if frozen_node_constraints <= frozen_parent_constraints:
                    include_node_schema = False
        if include_node_schema:
            schemas_to_check.append(node_schema)
        for schema in schemas_to_check:
            await self._check_one_schema(node=node, node_schema=schema, at=at, filters=filters)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from infrahub.core.constants import AttributeDBNodeType
from infrahub.core.constants.relationship_label import RELATIONSHIP_TO_NODE_LABEL, RELATIONSHIP_TO_VALUE_LABEL
from infrahub.core.constants.schema import FlagProperty, NodeProperty
from infrahub.core.query import Query, QueryNode, QueryRel, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.core.utils import build_regex_attrs

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.query import QueryElement
    from infrahub.database import InfrahubDatabase


class AttributeQuery(Query):
    def __init__(
        self,
        attr: BaseAttribute,
        attr_id: Optional[str] = None,
        at: Optional[Union[Timestamp, str]] = None,
        branch: Optional[Branch] = None,
        **kwargs: Any,
    ):
        self.attr = attr
        self.attr_id = attr_id or attr.db_id

        if at:
            self.at = Timestamp(at)
        else:
            self.at = self.attr.at

        self.branch = branch or self.attr.get_branch_based_on_support_type()

        super().__init__(**kwargs)


class AttributeUpdateValueQuery(AttributeQuery):
    name = "attribute_update_value"
    type: QueryType = QueryType.WRITE

    raise_error_if_empty: bool = True

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        at = self.at or self.attr.at

        self.params["attr_uuid"] = self.attr.id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        content = self.attr.to_db()
        self.params.update(self.attr.to_db())

        prop_list = [f"{key}: ${key}" for key in content.keys()]

        labels = ["AttributeValue"]
        node_type = self.attr.get_db_node_type()
        if node_type == AttributeDBNodeType.IPHOST:
            labels.append("AttributeIPHost")
        elif node_type == AttributeDBNodeType.IPNETWORK:
            labels.append("AttributeIPNetwork")

        query = """
        MATCH (a:Attribute { uuid: $attr_uuid })
        MERGE (av:%(labels)s { %(props)s } )
        WITH av, a
        LIMIT 1
        CREATE (a)-[r:%(rel_label)s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
        """ % {"rel_label": self.attr._rel_to_value_label, "labels": ":".join(labels), "props": ", ".join(prop_list)}

        self.add_to_query(query)
        self.return_labels = ["a", "av", "r"]


class AttributeUpdateFlagQuery(AttributeQuery):
    name = "attribute_update_flag"
    type: QueryType = QueryType.WRITE

    raise_error_if_empty: bool = True

    def __init__(
        self,
        flag_name: str,
        **kwargs: Any,
    ) -> None:
        SUPPORTED_FLAGS = ["is_visible", "is_protected"]

        if flag_name not in SUPPORTED_FLAGS:
            raise ValueError(f"Only {SUPPORTED_FLAGS} are supported for now.")

        self.flag_name = flag_name

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        at = self.at or self.attr.at

        self.params["attr_uuid"] = self.attr.id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        self.params["flag_value"] = getattr(self.attr, self.flag_name)
        self.params["flag_type"] = self.attr.get_kind()

        query = """
        MATCH (a:Attribute { uuid: $attr_uuid })
        MERGE (flag:Boolean { value: $flag_value })
        CREATE (a)-[r:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(flag)
        """ % self.flag_name.upper()

        self.add_to_query(query)
        self.return_labels = ["a", "flag", "r"]


class AttributeUpdateNodePropertyQuery(AttributeQuery):
    name = "attribute_update_node_property"
    type: QueryType = QueryType.WRITE

    raise_error_if_empty: bool = True

    def __init__(
        self,
        prop_name: str,
        prop_id: str,
        **kwargs: Any,
    ):
        self.prop_name = prop_name
        self.prop_id = prop_id

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        at = self.at or self.attr.at

        self.params["attr_uuid"] = self.attr.id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        self.params["prop_name"] = self.prop_name
        self.params["prop_id"] = self.prop_id

        rel_name = f"HAS_{self.prop_name.upper()}"

        query = (
            """
        MATCH (a:Attribute { uuid: $attr_uuid })
        MATCH (np:Node { uuid: $prop_id })
        CREATE (a)-[r:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(np)
        """
            % rel_name
        )

        self.add_to_query(query)
        self.return_labels = ["a", "np", "r"]


class AttributeGetQuery(AttributeQuery):
    name = "attribute_get"
    type: QueryType = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        self.params["attr_uuid"] = self.attr.id
        self.params["node_uuid"] = self.attr.node.id

        at = self.at or self.attr.at
        self.params["at"] = at.to_string()

        rels_filter, rels_params = self.branch.get_query_filter_path(at=at.to_string())
        self.params.update(rels_params)

        query = (
            """
        MATCH (a:Attribute { uuid: $attr_uuid })
        MATCH p = ((a)-[r2:HAS_VALUE|IS_VISIBLE|IS_PROTECTED|HAS_SOURCE|HAS_OWNER]->(ap))
        WHERE all(r IN relationships(p) WHERE ( %s ))
        """
            % rels_filter
        )

        self.add_to_query(query)

        self.return_labels = ["a", "ap", "r2"]


async def default_attribute_query_filter(  # pylint: disable=unused-argument,too-many-branches,too-many-statements
    name: str,
    filter_name: str,
    branch: Branch | None = None,
    filter_value: str | int | bool | list | None = None,
    attribute_kind: str | None = None,
    include_match: bool = True,
    param_prefix: str | None = None,
    db: InfrahubDatabase | None = None,
    partial_match: bool = False,
    support_profiles: bool = False,
) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
    """Generate Query String Snippet to filter the right node."""

    query_filter: list[QueryElement] = []
    query_params: dict[str, Any] = {}
    query_where: list[str] = []

    if filter_value and not isinstance(filter_value, (str, bool, int, list)):
        raise TypeError(f"filter {filter_name}: {filter_value} ({type(filter_value)}) is not supported.")

    if isinstance(filter_value, list) and not all(isinstance(value, (str, bool, int)) for value in filter_value):
        raise TypeError(f"filter {filter_name}: {filter_value} (list) contains unsupported item")

    param_prefix = param_prefix or f"attr_{name}"

    if include_match:
        query_filter.append(QueryNode(name="n"))

    query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_NODE_LABEL]))

    if name in ["any", "attribute"]:
        query_filter.append(QueryNode(name="i", labels=["Attribute"]))
    else:
        query_filter.append(QueryNode(name="i", labels=["Attribute"], params={"name": f"${param_prefix}_name"}))
        query_params[f"{param_prefix}_name"] = name

    if filter_name in ("value", "binary_address", "prefixlen", "isnull"):
        query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]))

        if filter_value is None:
            query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
        else:
            if partial_match:
                query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
                query_where.append(
                    f"toLower(toString(av.{filter_name})) CONTAINS toLower(toString(${param_prefix}_{filter_name}))"
                )
            elif attribute_kind and attribute_kind == "List" and not isinstance(filter_value, list):
                query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
                filter_value = build_regex_attrs(values=[filter_value])
                query_where.append(f"toString(av.{filter_name}) =~ ${param_prefix}_{filter_name}")
            elif filter_name == "isnull":
                query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
            elif support_profiles:
                query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
                query_where.append(f"(av.{filter_name} = ${param_prefix}_{filter_name} OR av.is_default)")
            else:
                query_filter.append(
                    QueryNode(
                        name="av", labels=["AttributeValue"], params={filter_name: f"${param_prefix}_{filter_name}"}
                    )
                )
            query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif filter_name == "values" and isinstance(filter_value, list):
        query_filter.extend(
            (QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]), QueryNode(name="av", labels=["AttributeValue"]))
        )
        if attribute_kind and attribute_kind == "List":
            query_params[f"{param_prefix}_{filter_name}"] = build_regex_attrs(values=filter_value)
            query_where.append(f"toString(av.value) =~ ${param_prefix}_{filter_name}")
        elif support_profiles:
            query_where.append(f"(av.value IN ${param_prefix}_value OR av.is_default)")
        else:
            query_where.append(f"av.value IN ${param_prefix}_value")
        query_params[f"{param_prefix}_value"] = filter_value

    elif filter_name == "version":
        query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]))

        if filter_value is None:
            query_filter.append(QueryNode(name="av", labels=["AttributeValue"]))
        else:
            query_filter.append(
                QueryNode(name="av", labels=["AttributeValue"], params={filter_name: f"${param_prefix}_{filter_name}"})
            )
            query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif filter_name in [v.value for v in FlagProperty] and filter_value is not None:
        query_filter.append(QueryRel(labels=[filter_name.upper()]))
        query_filter.append(
            QueryNode(name="ap", labels=["Boolean"], params={"value": f"${param_prefix}_{filter_name}"})
        )
        query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif "__" in filter_name and filter_value is not None:
        filter_name_split = filter_name.split(sep="__", maxsplit=1)
        property_name: str = filter_name_split[0]
        property_attr: str = filter_name_split[1]

        if property_name not in [v.value for v in NodeProperty]:
            raise ValueError(f"filter {filter_name}: {filter_value}, {property_name} is not a valid property")

        if property_attr not in ["id"]:
            raise ValueError(f"filter {filter_name}: {filter_value}, {property_attr} is supported")

        clean_filter_name = f"{property_name}_{property_attr}"

        query_filter.extend(
            [
                QueryRel(labels=[f"HAS_{property_name.upper()}"]),
                QueryNode(name="ap", labels=["Node"], params={"uuid": f"${param_prefix}_{clean_filter_name}"}),
            ]
        )
        query_params[f"{param_prefix}_{clean_filter_name}"] = filter_value

    return query_filter, query_params, query_where



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING, Any, Optional

from infrahub_sdk.utils import compare_lists, deep_merge_dict, duplicates, intersection
from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self

from infrahub.core.constants import (
    HashableModelState,
    SchemaPathType,
    UpdateSupport,
    UpdateValidationErrorType,
)
from infrahub.core.path import SchemaPath

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.schema_branch import SchemaBranch


class NodeKind(BaseModel):
    namespace: str
    name: str

    def __str__(self) -> str:
        return f"{self.namespace}{self.name}"


class SchemaBranchDiff(BaseModel):
    added_nodes: list[str] = Field(default_factory=list)
    changed_nodes: list[str] = Field(default_factory=list)
    added_generics: list[str] = Field(default_factory=list)
    changed_generics: list[str] = Field(default_factory=list)
    removed_nodes: list[str] = Field(default_factory=list)
    removed_generics: list[str] = Field(default_factory=list)

    def to_string(self) -> str:
        return ", ".join(self.nodes + self.generics)

    def to_list(self) -> list[str]:
        return self.nodes + self.generics

    @property
    def has_diff(self) -> bool:
        return any([self.has_node_diff, self.has_generic_diff])

    @property
    def has_node_diff(self) -> bool:
        return bool(self.added_nodes + self.changed_nodes + self.removed_nodes)

    @property
    def has_generic_diff(self) -> bool:
        return bool(self.added_generics + self.changed_generics + self.removed_generics)

    @property
    def nodes(self) -> list[str]:
        """Return nodes that are still active."""
        return self.added_nodes + self.changed_nodes

    @property
    def generics(self) -> list[str]:
        """Return generics that are still active."""
        return self.added_generics + self.changed_generics


class SchemaBranchHash(BaseModel):
    main: str
    nodes: dict[str, str] = Field(default_factory=dict)
    generics: dict[str, str] = Field(default_factory=dict)

    def compare(self, other: SchemaBranchHash) -> SchemaBranchDiff | None:
        if other.main == self.main:
            return None

        return SchemaBranchDiff(
            added_nodes=[key for key in other.nodes if key not in self.nodes],
            changed_nodes=[key for key, value in other.nodes.items() if key in self.nodes and self.nodes[key] != value],
            removed_nodes=[key for key in self.nodes if key not in other.nodes],
            added_generics=[key for key in other.generics if key not in self.generics],
            changed_generics=[
                key for key, value in other.generics.items() if key in self.generics and self.generics[key] != value
            ],
            removed_generics=[key for key in self.generics if key not in other.generics],
        )


class SchemaDiff(BaseModel):
    model_config = ConfigDict(extra="forbid")
    added: dict[str, HashableModelDiff] = Field(default_factory=dict)
    changed: dict[str, HashableModelDiff] = Field(default_factory=dict)
    removed: dict[str, HashableModelDiff] = Field(default_factory=dict)

    @property
    def all(self) -> list[str]:
        return list(self.changed.keys()) + list(self.added.keys()) + list(self.removed.keys())

    def __add__(self, other: SchemaDiff) -> SchemaDiff:
        merged_dict = deep_merge_dict(self.model_dump(), other.model_dump())
        return self.__class__(**merged_dict)

    def print(self, indentation: int = 4, column_size: int = 32) -> None:
        data = self.model_dump()

        indent_str = " " * indentation

        # pylint: disable=too-many-nested-blocks
        for node_action, node_info in data.items():
            for node_name, elements in node_info.items():
                print(f"{str(node_name).ljust(column_size)} | {str(node_action).title()}")
                for element_action, element_info in elements.items():
                    for element_name, element_children in element_info.items():
                        print(
                            f"{indent_str}{str(element_name).ljust(column_size - indentation)} | {str(element_action).title()}"
                        )
                        if element_children and isinstance(element_children, dict):
                            for sub_action, sub_info in element_children.items():
                                for sub_name in sub_info.keys():
                                    print(
                                        f"{indent_str * 2}{str(sub_name).ljust(column_size - indentation * 2)} | {str(sub_action).title()}"
                                    )


class SchemaUpdateValidationError(BaseModel):
    model_config = ConfigDict(extra="forbid")
    path: SchemaPath
    error: UpdateValidationErrorType
    message: Optional[str] = None

    def to_string(self) -> str:
        return f"{self.error.value!r}: {self.path.schema_kind} {self.path.field_name} {self.message}"


class SchemaUpdateMigrationInfo(BaseModel):
    model_config = ConfigDict(extra="forbid")
    path: SchemaPath
    migration_name: str

    @property
    def routing_key(self) -> str:
        return "schema.migration.path"


class SchemaUpdateConstraintInfo(BaseModel):
    model_config = ConfigDict(extra="forbid")
    path: SchemaPath
    constraint_name: str

    @property
    def routing_key(self) -> str:
        return "schema.validator.path"

    def __hash__(self) -> int:
        return hash((type(self),) + tuple([self.constraint_name + self.path.get_path()]))


class SchemaUpdateValidationResult(BaseModel):
    errors: list[SchemaUpdateValidationError] = Field(default_factory=list)
    constraints: list[SchemaUpdateConstraintInfo] = Field(default_factory=list)
    migrations: list[SchemaUpdateMigrationInfo] = Field(default_factory=list)
    enforce_update_support: bool = True
    diff: SchemaDiff

    @classmethod
    def init(cls, diff: SchemaDiff, schema: SchemaBranch, enforce_update_support: bool = True) -> Self:
        obj = cls(diff=diff, enforce_update_support=enforce_update_support)
        obj.process_diff(schema=schema)

        return obj

    def process_diff(self, schema: SchemaBranch) -> None:
        for schema_name, schema_diff in self.diff.removed.items():
            self.migrations.append(
                SchemaUpdateMigrationInfo(
                    path=SchemaPath(  # type: ignore[call-arg]
                        schema_kind=schema_name, path_type=SchemaPathType.NODE
                    ),
                    migration_name="node.remove",
                )
            )

        for schema_name, schema_diff in self.diff.changed.items():
            schema_node = schema.get(name=schema_name, duplicate=False)

            # Nothing to do today if we add a new attribute to a node in the schema
            # for node_field_name, _ in schema_diff.added.items():
            #     pass

            # Not possible today, removing an attribute from the schema will be manage by database migrations
            # for node_field_name, _ in schema_diff.removed.items():
            #     pass

            for node_field_name, node_field_diff in schema_diff.changed.items():
                if node_field_diff and node_field_name in ["attributes", "relationships"]:
                    self._process_attrs_rels(
                        schema=schema_node, node_field_name=node_field_name, node_field_diff=node_field_diff
                    )
                else:
                    self._process_node_attributes(schema=schema_node, node_field_name=node_field_name)

    def _process_attrs_rels(
        self,
        schema: MainSchemaTypes,
        node_field_name: str,
        node_field_diff: HashableModelDiff,
    ) -> None:
        path_type = SchemaPathType.ATTRIBUTE if node_field_name == "attributes" else SchemaPathType.RELATIONSHIP
        for field_name in node_field_diff.added.keys():
            if path_type == SchemaPathType.ATTRIBUTE:
                self.migrations.append(
                    SchemaUpdateMigrationInfo(
                        path=SchemaPath(  # type: ignore[call-arg]
                            schema_kind=schema.kind, path_type=path_type, field_name=field_name
                        ),
                        migration_name="node.attribute.add",
                    )
                )
            elif path_type == SchemaPathType.RELATIONSHIP:
                self.constraints.append(
                    SchemaUpdateConstraintInfo(
                        path=SchemaPath(  # type: ignore[call-arg]
                            schema_kind=schema.kind, path_type=path_type, field_name=field_name
                        ),
                        constraint_name="node.relationship.add",
                    )
                )

        for field_name in node_field_diff.removed.keys():
            self.migrations.append(
                SchemaUpdateMigrationInfo(  # type: ignore[call-arg]
                    path=SchemaPath(  # type: ignore[call-arg]
                        schema_kind=schema.kind, path_type=path_type, field_name=field_name
                    ),
                    migration_name=f"node.{path_type.value}.remove",
                )
            )

        for field_name, sub_field_diff in node_field_diff.changed.items():
            field = schema.get_field(name=field_name)

            if not sub_field_diff:
                raise ValueError("sub_field_diff must be defined, unexpected situation")

            for prop_name in sub_field_diff.changed:
                field_info = field.model_fields[prop_name]
                field_update = str(field_info.json_schema_extra.get("update"))  # type: ignore[union-attr]

                schema_path = SchemaPath(  # type: ignore[call-arg]
                    schema_kind=schema.kind,
                    path_type=path_type,
                    field_name=field_name,
                    property_name=prop_name,
                )

                self._process_field(
                    schema_path=schema_path,
                    field_update=field_update,
                )

    def _process_node_attributes(self, schema: MainSchemaTypes, node_field_name: str) -> None:
        field_info = schema.model_fields[node_field_name]
        field_update = str(field_info.json_schema_extra.get("update"))  # type: ignore[union-attr]

        schema_path = SchemaPath(  # type: ignore[call-arg]
            schema_kind=schema.kind,
            path_type=SchemaPathType.NODE,
            field_name=node_field_name,
            property_name=node_field_name,
        )
        self._process_field(
            schema_path=schema_path,
            field_update=field_update,
        )

    def _process_field(
        self,
        schema_path: SchemaPath,
        field_update: str,
    ) -> None:
        if field_update == UpdateSupport.NOT_SUPPORTED.value and self.enforce_update_support:
            self.errors.append(
                SchemaUpdateValidationError(
                    path=schema_path,
                    error=UpdateValidationErrorType.NOT_SUPPORTED,
                )
            )
        elif field_update == UpdateSupport.MIGRATION_REQUIRED.value:
            migration_name = f"{schema_path.path_type.value}.{schema_path.property_name}.update"
            self.migrations.append(
                SchemaUpdateMigrationInfo(
                    path=schema_path,
                    migration_name=migration_name,
                )
            )
        elif field_update == UpdateSupport.VALIDATE_CONSTRAINT.value:
            constraint_name = f"{schema_path.path_type.value}.{schema_path.property_name}.update"
            self.constraints.append(
                SchemaUpdateConstraintInfo(
                    path=schema_path,
                    constraint_name=constraint_name,
                )
            )

    def validate_all(self, migration_map: dict[str, Any], validator_map: dict[str, Any]) -> None:
        self.validate_migrations(migration_map=migration_map)
        self.validate_constraints(validator_map=validator_map)
        self.add_validator_for_migration(validator_map=validator_map)

    def validate_migrations(self, migration_map: dict[str, Any]) -> None:
        for migration in self.migrations:
            if migration_map.get(migration.migration_name, None) is None:
                self.errors.append(
                    SchemaUpdateValidationError(
                        path=migration.path,
                        error=UpdateValidationErrorType.MIGRATION_NOT_AVAILABLE,
                        message=f"Migration {migration.migration_name!r} is not available yet",
                    )
                )

    def validate_constraints(self, validator_map: dict[str, Any]) -> None:
        for constraint in self.constraints:
            if validator_map.get(constraint.constraint_name, None) is None:
                self.errors.append(
                    SchemaUpdateValidationError(
                        path=constraint.path,
                        error=UpdateValidationErrorType.VALIDATOR_NOT_AVAILABLE,
                        message=f"Validator {constraint.constraint_name!r} is not available yet",
                    )
                )

    def add_validator_for_migration(self, validator_map: dict[str, Any]) -> None:
        for migration in self.migrations:
            if validator_map.get(migration.migration_name, None):
                self.constraints.append(
                    SchemaUpdateConstraintInfo(
                        path=migration.path,
                        constraint_name=migration.migration_name,
                    )
                )


class HashableModelDiff(BaseModel):
    model_config = ConfigDict(extra="forbid")
    added: dict[str, Optional[HashableModelDiff]] = Field(default_factory=dict)
    changed: dict[str, Optional[HashableModelDiff]] = Field(default_factory=dict)
    removed: dict[str, Optional[HashableModelDiff]] = Field(default_factory=dict)

    @property
    def has_diff(self) -> bool:
        return bool(len(self.added) + len(self.changed) + len(self.removed))


class HashableModel(BaseModel):
    model_config = ConfigDict(extra="forbid")

    id: Optional[str] = None
    state: HashableModelState = HashableModelState.PRESENT

    _exclude_from_hash: list[str] = []
    _sort_by: list[str] = []

    def __hash__(self) -> int:
        return hash(self.get_hash())

    def get_hash(self, display_values: bool = False) -> str:
        """Generate a hash for the object.

        Calculating a hash can be very complicated if the data that we are storing is dynamic
        In order for this function to work, it's recommended to exclude all objects or list of objects with _exclude_from_hash
        List of hashable elements are fine and they will be converted automatically to Tuple.
        """

        values = []
        md5hash = hashlib.md5(usedforsecurity=False)
        for field_name in sorted(self.model_fields.keys()):
            if field_name.startswith("_") or field_name in self._exclude_from_hash:
                continue

            value = getattr(self, field_name)
            signatures = self._get_signature_field(value)
            for item in signatures:
                values.append(item)
                md5hash.update(item)

        if display_values:
            from rich import print as rprint  # pylint: disable=import-outside-toplevel

            rprint(tuple(values))

        return md5hash.hexdigest()

    @classmethod
    def _get_hash_value(cls, value: Any) -> bytes:
        if hasattr(value, "get_hash"):
            return value.get_hash().encode()

        return str(value).encode()

    @classmethod
    def _get_signature_field(cls, value: Any) -> list[bytes]:
        hashes: list[bytes] = []
        if isinstance(value, list):
            for item in sorted(value):
                hashes.append(cls._get_hash_value(item))
        elif isinstance(value, dict):
            for key, content in frozenset(sorted(value.items())):
                hashes.append(cls._get_hash_value(key))
                hashes.append(cls._get_hash_value(content))
        else:
            hashes.append(cls._get_hash_value(value))

        return sorted(hashes)

    @property
    def _sorting_id(self) -> tuple[Any]:
        return tuple(getattr(self, key) for key in self._sort_by if hasattr(self, key))

    def _sorting_keys(self, other: HashableModel) -> tuple[list[Any], list[Any]]:
        """Retrieve the values of the attributes listed in the _sort_key list, for both objects."""
        if not self._sort_by:
            raise TypeError(f"Sorting not supported for instance of {self.__class__.__name__}")

        if not hasattr(other, "_sort_by") and not other._sort_by:
            raise TypeError(
                f"Sorting not supported between instance of {other.__class__.__name__} and {self.__class__.__name__}"
            )

        self_sort_keys: list[Any] = [getattr(self, key) for key in self._sort_by if hasattr(self, key)]
        other_sort_keys: list[Any] = [getattr(other, key) for key in other._sort_by if hasattr(other, key)]

        return self_sort_keys, other_sort_keys

    def __lt__(self, other: Self) -> bool:
        self_sort_keys, other_sort_keys = self._sorting_keys(other)
        return tuple(self_sort_keys) < tuple(other_sort_keys)

    def __le__(self, other: Self) -> bool:
        self_sort_keys, other_sort_keys = self._sorting_keys(other)
        return tuple(self_sort_keys) <= tuple(other_sort_keys)

    def __gt__(self, other: Self) -> bool:
        self_sort_keys, other_sort_keys = self._sorting_keys(other)
        return tuple(self_sort_keys) > tuple(other_sort_keys)

    def __ge__(self, other: Self) -> bool:
        self_sort_keys, other_sort_keys = self._sorting_keys(other)
        return tuple(self_sort_keys) >= tuple(other_sort_keys)

    def duplicate(self) -> Self:
        """Duplicate the current object by doing a deep copy of everything and recreating a new object."""
        return self.model_copy(deep=True)

    @staticmethod
    def is_list_composed_of_hashable_model(items: list[Any]) -> bool:
        return all(isinstance(item, HashableModel) for item in items)

    @staticmethod
    def _organize_sub_items(items: list[HashableModel], shared_ids: set[str]) -> dict[tuple[Any], HashableModel]:
        """Convert a list of HashableModel into a dict with the sorting_id is the key, or the id if it was provided as part of the shared_ids"""
        sub_items = {}
        for item in items:
            if item.id and item.id in shared_ids:
                sub_items[item.id,] = item
                continue
            sub_items[item._sorting_id] = item

        return sub_items

    @staticmethod
    def update_list_hashable_model(
        field_name: str, attr_local: list[HashableModel], attr_other: list[HashableModel]
    ) -> list[Any]:
        """
        Merging the list is not easy,
        we need to create a unique id based on the sorting keys
        and if we have 2 sub items with the same key we can merge them recursively with update()
        """

        # Identify all nodes that are sharing a real IDs
        local_sub_real_ids = {item.id for item in attr_local if item.id}
        other_sub_real_ids = {item.id for item in attr_other if item.id}
        shared_real_ids = local_sub_real_ids & other_sub_real_ids

        local_sub_items = HashableModel._organize_sub_items(items=attr_local, shared_ids=shared_real_ids)
        other_sub_items = HashableModel._organize_sub_items(items=attr_other, shared_ids=shared_real_ids)

        new_list = []

        if len(local_sub_items) != len(attr_local) or len(other_sub_items) != len(attr_other):
            raise ValueError(f"Unable to merge the list for {field_name}, not all items are supporting _sorting_id")

        if duplicates(list(local_sub_items.keys())) or duplicates(list(other_sub_items.keys())):
            raise ValueError(f"Unable to merge the list for {field_name}, some items have the same _sorting_id")

        shared_ids = intersection(list(local_sub_items.keys()), list(other_sub_items.keys()))
        local_only_ids = set(list(local_sub_items.keys())) - set(shared_ids)
        other_only_ids = set(list(other_sub_items.keys())) - set(shared_ids)

        new_list = [value for key, value in local_sub_items.items() if key in local_only_ids]
        new_list.extend(
            [
                value
                for key, value in other_sub_items.items()
                if key in other_only_ids and value.state != HashableModelState.ABSENT
            ]
        )

        for item_id in shared_ids:
            other_item = other_sub_items[item_id]
            local_item = local_sub_items[item_id]
            if other_item.state == HashableModelState.ABSENT:
                continue
            new_list.append(local_item.update(other_item))

        return new_list

    def update(self, other: Self) -> Self:
        """Update the current object with the new value from the new one if they are defined.

        Currently this method works for the following type of fields
            int, str, bool, float: If present the value from Other is overwriting the local value
            list[BaseSchemaModel]: The list will be merge if all elements in the list have a _sorting_id and if it's unique.
            HashableModel: the value will be merged using update()

        TODO Implement other fields type like dict
        """

        for field_name in other.model_fields.keys():
            if not hasattr(self, field_name):
                setattr(self, field_name, getattr(other, field_name))
                continue

            attr_other = getattr(other, field_name)
            attr_local = getattr(self, field_name)
            if attr_other is None or attr_local == attr_other:
                continue

            if attr_local is None or isinstance(attr_other, (int, str, bool, float)):
                setattr(self, field_name, attr_other)
                continue

            if isinstance(attr_local, list) and isinstance(attr_other, list):
                new_attr = self._get_updated_list_value(
                    field_name=field_name, attr_local=attr_local, attr_other=attr_other
                )
                setattr(self, field_name, new_attr)
                continue

            if isinstance(attr_local, HashableModel) and isinstance(attr_other, HashableModel):
                attr_local.update(attr_other)
                continue

        return self

    def _get_updated_list_value(self, field_name: str, attr_local: list[Any], attr_other: list[Any]) -> list[Any]:
        if self.is_list_composed_of_hashable_model(attr_local) and self.is_list_composed_of_hashable_model(attr_other):
            return self.update_list_hashable_model(field_name=field_name, attr_local=attr_local, attr_other=attr_other)

        return attr_other

    def _get_field_names_for_diff(self) -> list[str]:
        return list(self.model_fields.keys())

    def diff(self, other: Self) -> HashableModelDiff:
        in_both, local_only, other_only = compare_lists(
            list1=self._get_field_names_for_diff(), list2=other._get_field_names_for_diff()
        )
        diff_result = HashableModelDiff(added=dict.fromkeys(local_only), removed=dict.fromkeys(other_only))

        for field_name in in_both:
            if field_name.startswith("_") or field_name in self._exclude_from_hash:
                continue

            local_value = getattr(self, field_name)
            other_value = getattr(other, field_name)
            local_signatures = self._get_signature_field(local_value)
            other_signatures = other._get_signature_field(other_value)

            if local_signatures != other_signatures:
                if isinstance(local_value, HashableModel) and isinstance(other_value, HashableModel):
                    diff_result.changed[field_name] = local_value.diff(other=other_value)
                else:
                    diff_result.changed[field_name] = None

        return diff_result



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class CheckRepositoryMergeConflicts(InfrahubMessage):
    """Runs a check to validate if there are merge conflicts for a proposed change between two branches."""

    validator_id: str = Field(..., description="The id of the validator associated with this check")
    validator_execution_id: str = Field(..., description="The id of current execution of the associated validator")
    check_execution_id: str = Field(..., description="The unique ID for the current execution of this check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    source_branch: str = Field(..., description="The source branch")
    target_branch: str = Field(..., description="The target branch")



from pydantic import Field

from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.messages.refresh_registry_branches import RefreshRegistryBranches
from infrahub.message_bus.messages.refresh_registry_rebasedbranch import RefreshRegistryRebasedBranch

from .models import InfrahubBranchEvent


class BranchDeleteEvent(InfrahubBranchEvent):
    """Event generated when a branch has been deleted"""

    branch_id: str = Field(..., description="The ID of the mutated node")
    sync_with_git: bool = Field(..., description="Indicates if the branch was extended to Git")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.branch.deleted"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.branch.{self.branch}",
            "infrahub.branch.id": self.branch_id,
            "infrahub.branch.name": self.branch,
        }

    def get_messages(self) -> list[InfrahubMessage]:
        events: list[InfrahubMessage] = [
            # EventBranchDelete(
            #     branch=self.branch,
            #     branch_id=self.branch_id,
            #     sync_with_git=self.sync_with_git,
            #     meta=self.get_message_meta(),
            # ),
            RefreshRegistryBranches(),
        ]
        return events


class BranchCreateEvent(InfrahubBranchEvent):
    """Event generated when a branch has been created"""

    branch_id: str = Field(..., description="The ID of the mutated node")
    sync_with_git: bool = Field(..., description="Indicates if the branch was extended to Git")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.branch.created"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.branch.{self.branch}",
            "infrahub.branch.id": self.branch_id,
            "infrahub.branch.name": self.branch,
        }

    def get_messages(self) -> list[InfrahubMessage]:
        events: list[InfrahubMessage] = [
            # EventBranchCreate(
            #     branch=self.branch,
            #     branch_id=self.branch_id,
            #     sync_with_git=self.sync_with_git,
            #     meta=self.get_message_meta(),
            # ),
            RefreshRegistryBranches(),
        ]
        return events


class BranchRebaseEvent(InfrahubBranchEvent):
    """Event generated when a branch has been rebased"""

    branch_id: str = Field(..., description="The ID of the mutated node")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.branch.rebased"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.branch.{self.branch}",
            "infrahub.branch.id": self.branch_id,
        }

    def get_messages(self) -> list[InfrahubMessage]:
        events: list[InfrahubMessage] = [
            # EventBranchRebased(
            #     branch=self.branch,
            #     meta=self.get_message_meta(),
            # ),
            RefreshRegistryRebasedBranch(branch=self.branch),
        ]
        return events



from __future__ import annotations

from infrahub_sdk.batch import InfrahubBatch
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.path import SchemaPath  # noqa: TC001
from infrahub.core.schema import GenericSchema, NodeSchema  # noqa: TC001
from infrahub.core.validators.aggregated_checker import AggregatedConstraintChecker
from infrahub.core.validators.model import (
    SchemaConstraintValidatorRequest,
)
from infrahub.dependencies.registry import get_component_registry
from infrahub.services import services
from infrahub.workflows.utils import add_tags

from .models.validate_migration import SchemaValidateMigrationData, SchemaValidatorPathResponseData


@flow(name="schema_validate_migrations", flow_run_name="Validate schema migrations", persist_result=True)
async def schema_validate_migrations(message: SchemaValidateMigrationData) -> list[SchemaValidatorPathResponseData]:
    batch = InfrahubBatch(return_exceptions=True)
    log = get_run_logger()
    await add_tags(branches=[message.branch.name])

    if not message.constraints:
        log.info("No constaint to validate")
        return []

    log.info(f"{len(message.constraints)} constraint(s) to validate")
    # NOTE this task is a good candidate to add a progress bar
    for constraint in message.constraints:
        schema = message.schema_branch.get(name=constraint.path.schema_kind)
        if not isinstance(schema, (GenericSchema, NodeSchema)):
            continue
        batch.add(
            task=schema_path_validate,
            branch=message.branch,
            constraint_name=constraint.constraint_name,
            node_schema=schema,
            schema_path=constraint.path,
        )

    results = [result async for _, result in batch.execute()]
    return results


@task(
    name="schema-path-validate",
    task_run_name="Validate schema path {constraint_name} in {branch.name}",
    description="Validate if a given migration is compatible with the existing data",
    retries=3,
    cache_policy=NONE,
)
async def schema_path_validate(
    branch: Branch,
    constraint_name: str,
    node_schema: NodeSchema | GenericSchema,
    schema_path: SchemaPath,
) -> SchemaValidatorPathResponseData:
    service = services.service

    async with service.database.start_session() as db:
        constraint_request = SchemaConstraintValidatorRequest(
            branch=branch,
            constraint_name=constraint_name,
            node_schema=node_schema,
            schema_path=schema_path,
        )

        component_registry = get_component_registry()
        aggregated_constraint_checker = await component_registry.get_component(
            AggregatedConstraintChecker, db=db, branch=branch
        )
        violations = await aggregated_constraint_checker.run_constraints(constraint_request)

        return SchemaValidatorPathResponseData(
            violations=violations, constraint_name=constraint_name, schema_path=schema_path
        )



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class RefreshGitFetch(InfrahubMessage):
    """Fetch a repository remote changes."""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The type of repository")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Id of the Infrahub branch on which to sync the remote repository")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from graphene import InputObjectType, Mutation
from graphene.types.mutation import MutationOptions
from infrahub_sdk.utils import extract_fields
from typing_extensions import Self

from infrahub import config, lock
from infrahub.core import registry
from infrahub.core.constants import InfrahubKind, MutationAction
from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.profile_schema import ProfileSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import retry_db_transaction
from infrahub.dependencies.registry import get_component_registry
from infrahub.events import EventMeta, NodeMutatedEvent
from infrahub.exceptions import ValidationError
from infrahub.lock import InfrahubMultiLock, build_object_lock_name
from infrahub.log import get_log_data, get_logger
from infrahub.worker import WORKER_IDENTITY

from .node_getter.by_default_filter import MutationNodeGetterByDefaultFilter
from .node_getter.by_hfid import MutationNodeGetterByHfid
from .node_getter.by_id import MutationNodeGetterById

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase

    from ..initialization import GraphqlContext
    from .node_getter.interface import MutationNodeGetterInterface

# pylint: disable=unused-argument

log = get_logger()

KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED = [InfrahubKind.GENERICGROUP]


# ------------------------------------------
# Infrahub GraphQLType
# ------------------------------------------
class InfrahubMutationOptions(MutationOptions):
    schema: Optional[NodeSchema] = None


class InfrahubMutationMixin:
    @classmethod
    async def mutate(cls, root: dict, info: GraphQLResolveInfo, data: InputObjectType, *args: Any, **kwargs):
        context: GraphqlContext = info.context

        obj = None
        mutation = None
        action = MutationAction.UNDEFINED

        if "Create" in cls.__name__:
            obj, mutation = await cls.mutate_create(info=info, branch=context.branch, data=data, **kwargs)
            action = MutationAction.ADDED
        elif "Update" in cls.__name__:
            obj, mutation = await cls.mutate_update(info=info, branch=context.branch, data=data, **kwargs)
            action = MutationAction.UPDATED
        elif "Upsert" in cls.__name__:
            node_manager = NodeManager()
            node_getters = [
                MutationNodeGetterById(db=context.db, node_manager=node_manager),
                MutationNodeGetterByHfid(db=context.db, node_manager=node_manager),
                MutationNodeGetterByDefaultFilter(db=context.db, node_manager=node_manager),
            ]
            obj, mutation, created = await cls.mutate_upsert(
                info=info, branch=context.branch, data=data, node_getters=node_getters, **kwargs
            )
            if created:
                action = MutationAction.ADDED
            else:
                action = MutationAction.UPDATED
        elif "Delete" in cls.__name__:
            obj, mutation = await cls.mutate_delete(info=info, branch=context.branch, data=data, **kwargs)
            action = MutationAction.REMOVED
        else:
            raise ValueError(
                f"Unexpected class Name: {cls.__name__}, should end with Create, Update, Upsert, or Delete"
            )

        # Reset the time of the query to guarantee that all resolvers executed after this point will account for the changes
        context.at = Timestamp()

        if config.SETTINGS.broker.enable and context.background:
            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            graphql_payload = await obj.to_graphql(db=context.db, filter_sensitive=True, include_properties=False)
            event = NodeMutatedEvent(
                branch=context.branch.name,
                kind=obj._schema.kind,
                node_id=obj.id,
                data=graphql_payload,
                action=action,
                fields=_get_data_fields(data),
                meta=EventMeta(initiator_id=WORKER_IDENTITY, request_id=request_id),
            )

            context.background.add_task(context.active_service.event.send, event)

        return mutation

    @classmethod
    async def _get_profile_ids(cls, db: InfrahubDatabase, obj: Node) -> set[str]:
        if not hasattr(obj, "profiles"):
            return set()
        profile_rels = await obj.profiles.get_relationships(db=db)
        return {pr.peer_id for pr in profile_rels}

    @classmethod
    async def _refresh_for_profile_update(
        cls, db: InfrahubDatabase, branch: Branch, obj: Node, previous_profile_ids: Optional[set[str]] = None
    ) -> Node:
        if not hasattr(obj, "profiles"):
            return obj
        current_profile_ids = await cls._get_profile_ids(db=db, obj=obj)
        if previous_profile_ids is None or previous_profile_ids != current_profile_ids:
            return await NodeManager.get_one_by_id_or_default_filter(
                db=db,
                kind=cls._meta.schema.kind,
                id=obj.get_id(),
                branch=branch,
                include_owner=True,
                include_source=True,
            )
        return obj

    @classmethod
    async def _call_mutate_create_object(cls, data: InputObjectType, db: InfrahubDatabase, branch: Branch):
        """
        Wrapper around mutate_create_object to potentially activate locking.
        """
        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = _get_kind_lock_names_on_object_mutation(
            kind=cls._meta.schema.kind, branch=branch, schema_branch=schema_branch
        )
        if lock_names:
            async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                return await cls.mutate_create_object(data=data, db=db, branch=branch)

        return await cls.mutate_create_object(data=data, db=db, branch=branch)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db
        obj = await cls._call_mutate_create_object(data=data, db=db, branch=branch)
        result = await cls.mutate_create_to_graphql(info=info, db=db, obj=obj)
        return obj, result

    @classmethod
    @retry_db_transaction(name="object_create")
    async def mutate_create_object(
        cls,
        data: InputObjectType,
        db: InfrahubDatabase,
        branch: Branch,
    ) -> Node:
        component_registry = get_component_registry()
        node_constraint_runner = await component_registry.get_component(
            NodeConstraintRunner, db=db.start_session(), branch=branch
        )
        node_class = Node
        if cls._meta.schema.kind in registry.node:
            node_class = registry.node[cls._meta.schema.kind]

        fields_to_validate = list(data)
        try:
            if db.is_transaction:
                obj = await node_class.init(db=db, schema=cls._meta.schema, branch=branch)
                await obj.new(db=db, **data)
                await node_constraint_runner.check(node=obj, field_filters=fields_to_validate)
                await obj.save(db=db)
            else:
                async with db.start_transaction() as dbt:
                    obj = await node_class.init(db=dbt, schema=cls._meta.schema, branch=branch)
                    await obj.new(db=dbt, **data)
                    await node_constraint_runner.check(node=obj, field_filters=fields_to_validate)
                    await obj.save(db=dbt)

        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        if await cls._get_profile_ids(db=db, obj=obj):
            obj = await cls._refresh_for_profile_update(db=db, branch=branch, obj=obj)

        return obj

    @classmethod
    async def mutate_create_to_graphql(cls, info: GraphQLResolveInfo, db: InfrahubDatabase, obj: Node) -> Self:
        fields = await extract_fields(info.field_nodes[0].selection_set)
        result = {"ok": True}
        if "object" in fields:
            result["object"] = await obj.to_graphql(db=db, fields=fields.get("object", {}))
        return cls(**result)

    @classmethod
    async def _call_mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        obj: Node,
    ) -> tuple[Node, Self]:
        """
        Wrapper around mutate_update to potentially activate locking and call it within a database transaction.
        """

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = _get_kind_lock_names_on_object_mutation(
            kind=cls._meta.schema.kind, branch=branch, schema_branch=schema_branch
        )

        if db.is_transaction:
            if lock_names:
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                    obj = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=obj)
            else:
                obj = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=obj)
            result = await cls.mutate_update_to_graphql(db=db, info=info, obj=obj)
            return obj, result

        async with db.start_transaction() as dbt:
            if lock_names:
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names):
                    obj = await cls.mutate_update_object(db=dbt, info=info, data=data, branch=branch, obj=obj)
            else:
                obj = await cls.mutate_update_object(db=dbt, info=info, data=data, branch=branch, obj=obj)
            result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=obj)
            return obj, result

    @classmethod
    @retry_db_transaction(name="object_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db

        obj = node or await NodeManager.find_object(
            db=db, kind=cls._meta.schema.kind, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )

        try:
            obj, result = await cls._call_mutate_update(info=info, data=data, db=db, branch=branch, obj=obj)
        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        return obj, result

    @classmethod
    async def mutate_update_object(
        cls, db: InfrahubDatabase, info: GraphQLResolveInfo, data: InputObjectType, branch: Branch, obj: Node
    ) -> Node:
        component_registry = get_component_registry()
        node_constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

        before_mutate_profile_ids = await cls._get_profile_ids(db=db, obj=obj)
        await obj.from_graphql(db=db, data=data)
        fields_to_validate = list(data)
        await node_constraint_runner.check(node=obj, field_filters=fields_to_validate)

        fields = list(data.keys())
        for field in ("id", "hfid"):
            if field in fields:
                fields.remove(field)

        await obj.save(db=db, fields=fields)
        obj = await cls._refresh_for_profile_update(
            db=db, branch=branch, obj=obj, previous_profile_ids=before_mutate_profile_ids
        )
        return obj

    @classmethod
    async def mutate_update_to_graphql(
        cls,
        db: InfrahubDatabase,
        info: GraphQLResolveInfo,
        obj: Node,
    ) -> Self:
        fields_object = await extract_fields(info.field_nodes[0].selection_set)
        fields_object = fields_object.get("object", {})
        result = {"ok": True}
        if fields_object:
            result["object"] = await obj.to_graphql(db=db, fields=fields_object)
        return cls(**result)

    @classmethod
    @retry_db_transaction(name="object_upsert")
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getters: list[MutationNodeGetterInterface],
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self, bool]:
        schema_name = cls._meta.schema.kind

        context: GraphqlContext = info.context
        db = database or context.db

        node_schema = db.schema.get(name=schema_name, branch=branch)

        node = None
        for getter in node_getters:
            node = await getter.get_node(node_schema=node_schema, data=data, branch=branch)
            if node:
                break

        if node:
            updated_obj, mutation = await cls.mutate_update(info=info, data=data, branch=branch, database=db, node=node)
            return updated_obj, mutation, False
        # We need to convert the InputObjectType into a dict in order to remove hfid that isn't a valid input when creating the object
        data_dict = dict(data)
        if "hfid" in data:
            del data_dict["hfid"]
        created_obj, mutation = await cls.mutate_create(info=info, data=data_dict, branch=branch)
        return created_obj, mutation, True

    @classmethod
    @retry_db_transaction(name="object_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        obj = await NodeManager.find_object(
            db=context.db, kind=cls._meta.schema.kind, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )

        try:
            async with context.db.start_transaction() as db:
                deleted = await NodeManager.delete(db=db, branch=branch, nodes=[obj])
        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        deleted_str = ", ".join([f"{d.get_kind()}({d.get_id()})" for d in deleted])
        log.info(f"nodes deleted: {deleted_str}")

        ok = True

        return obj, cls(ok=ok)


class InfrahubMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls, schema: Optional[Union[NodeSchema, GenericSchema, ProfileSchema]] = None, _meta=None, **options
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, (NodeSchema, GenericSchema, ProfileSchema)):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)


def _get_kinds_to_lock_on_object_mutation(kind: str, schema_branch: SchemaBranch) -> list[str]:
    """
    Return kinds for which we want to lock during creating / updating an object of a given schema node.
    Lock should be performed on schema kind and its generics having a uniqueness_constraint defined.
    If a generic uniqueness constraint is the same as the node schema one,
    it means node schema overrided this constraint, in which case we only need to lock on the generic.
    """

    node_schema = schema_branch.get(name=kind)

    schema_uc = None
    kinds = []
    if node_schema.uniqueness_constraints:
        kinds.append(node_schema.kind)
        schema_uc = node_schema.uniqueness_constraints

    if node_schema.is_generic_schema:
        return kinds

    generics_kinds = node_schema.inherit_from

    node_schema_kind_removed = False
    for generic_kind in generics_kinds:
        generic_uc = schema_branch.get(name=generic_kind).uniqueness_constraints
        if generic_uc:
            kinds.append(generic_kind)
            if not node_schema_kind_removed and generic_uc == schema_uc:
                # Check whether we should remove original schema kind as it simply overrides uniqueness_constraint
                # of a generic
                kinds.pop(0)
                node_schema_kind_removed = True
    return kinds


def _should_kind_be_locked_on_any_branch(kind: str, schema_branch: SchemaBranch) -> bool:
    """
    Check whether kind or any kind generic is in KINDS_TO_LOCK_ON_ANY_BRANCH.
    """

    if kind in KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED:
        return True

    node_schema = schema_branch.get(name=kind)
    if node_schema.is_generic_schema:
        return False

    for generic_kind in node_schema.inherit_from:
        if generic_kind in KINDS_CONCURRENT_MUTATIONS_NOT_ALLOWED:
            return True
    return False


def _get_kind_lock_names_on_object_mutation(kind: str, branch: Branch, schema_branch: SchemaBranch) -> list[str]:
    """
    Return objects kind for which we want to avoid concurrent mutation (create/update). Except for some specific kinds,
    concurrent mutations are only allowed on non-main branch as objects validations will be performed at least when merging in main branch.
    """

    if not branch.is_default and not _should_kind_be_locked_on_any_branch(kind, schema_branch):
        return []

    lock_kinds = _get_kinds_to_lock_on_object_mutation(kind, schema_branch)
    lock_names = [build_object_lock_name(kind) for kind in lock_kinds]
    return lock_names


def _get_data_fields(data: InputObjectType) -> list[str]:
    return [field for field in data.keys() if field not in ["id", "hfid"]]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from graphql import graphql

from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreGraphQLQuery
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.graphql.initialization import prepare_graphql_params

if TYPE_CHECKING:
    from graphql.execution import ExecutionResult

    from infrahub.database import InfrahubDatabase


async def execute_query(
    name: str,
    db: InfrahubDatabase,
    params: Optional[dict] = None,
    branch: Branch | str | None = None,
    at: Timestamp | str | None = None,
) -> ExecutionResult:
    """Helper function to Execute a GraphQL Query."""

    if not isinstance(branch, Branch):
        branch = await registry.get_branch(db=db, branch=branch)
    at = Timestamp(at)

    graphql_query = await NodeManager.get_one_by_default_filter(
        db=db, id=name, kind=CoreGraphQLQuery, branch=branch, at=at
    )
    if not graphql_query:
        raise ValueError(f"Unable to find the {InfrahubKind.GRAPHQLQUERY} {name}")

    gql_params = await prepare_graphql_params(
        branch=branch, db=db, at=at, include_mutation=False, include_subscription=False
    )

    result = await graphql(
        schema=gql_params.schema,
        source=graphql_query.query.value,
        context_value=gql_params.context,
        root_value=None,
        variable_values=params or {},
    )

    return result



from infrahub.git.directory import initialize_repositories_directory
from infrahub.git.repository import (
    InfrahubReadOnlyRepository,
    InfrahubRepository,
)

__all__ = [
    "InfrahubReadOnlyRepository",
    "InfrahubRepository",
    "initialize_repositories_directory",
]



from typing import Any

from infrahub.core.constants import (
    DEFAULT_KIND_MAX_LENGTH,
    DEFAULT_KIND_MIN_LENGTH,
    NAMESPACE_REGEX,
    AccountStatus,
    AccountType,
    AllowOverrideType,
    ArtifactStatus,
    BranchConflictKeep,
    BranchSupportType,
    ContentType,
    GeneratorInstanceStatus,
    GlobalPermissions,
    InfrahubKind,
    PermissionAction,
    PermissionDecision,
    RelationshipDeleteBehavior,
    RepositoryInternalStatus,
    RepositoryOperationalStatus,
    RepositorySyncStatus,
    Severity,
    ValidatorConclusion,
    ValidatorState,
)
from infrahub.proposed_change.constants import ProposedChangeState

# pylint: disable=too-many-lines

core_profile_schema_definition: dict[str, Any] = {
    "name": "Profile",
    "namespace": "Core",
    "include_in_menu": False,
    "icon": "mdi:shape-plus-outline",
    "description": "Base Profile in Infrahub.",
    "label": "Profile",
    "display_labels": ["profile_name__value"],
    "default_filter": "profile_name__value",
    "uniqueness_constraints": [["profile_name__value"]],
    "attributes": [
        {
            "name": "profile_name",
            "kind": "Text",
            "min_length": 3,
            "max_length": 32,
            "optional": False,
            "unique": True,
        },
        {
            "name": "profile_priority",
            "kind": "Number",
            "default_value": 1000,
            "optional": True,
        },
    ],
}

# -----------------------------------------------
# Menu Items
# -----------------------------------------------
generic_menu_item: dict[str, Any] = {
    "name": "Menu",
    "namespace": "Core",
    "include_in_menu": False,
    "description": "Element of the Menu",
    "label": "Menu",
    "hierarchical": True,
    "human_friendly_id": ["namespace__value", "name__value"],
    "display_labels": ["label__value"],
    "generate_profile": False,
    "attributes": [
        {"name": "namespace", "kind": "Text", "regex": NAMESPACE_REGEX, "order_weight": 1000},
        {"name": "name", "kind": "Text", "order_weight": 1000},
        {"name": "label", "kind": "Text", "optional": True, "order_weight": 2000},
        {"name": "kind", "kind": "Text", "optional": True, "order_weight": 2500},
        {"name": "path", "kind": "Text", "optional": True, "order_weight": 2500},
        {"name": "description", "kind": "Text", "optional": True, "order_weight": 3000},
        {"name": "icon", "kind": "Text", "optional": True, "order_weight": 4000},
        {"name": "protected", "kind": "Boolean", "default_value": False, "read_only": True, "order_weight": 5000},
        {"name": "order_weight", "kind": "Number", "default_value": 2000, "order_weight": 6000},
        {"name": "required_permissions", "kind": "List", "optional": True, "order_weight": 7000},
        {
            "name": "section",
            "kind": "Text",
            "enum": ["object", "internal"],
            "default_value": "object",
            "order_weight": 8000,
        },
    ],
}

menu_item: dict[str, Any] = {
    "name": "MenuItem",
    "namespace": "Core",
    "include_in_menu": False,
    "description": "Menu Item",
    "label": "Menu Item",
    "inherit_from": ["CoreMenu"],
    "generate_profile": False,
}

core_models: dict[str, Any] = {
    "generics": [
        {
            "name": "Node",
            "namespace": "Core",
            "include_in_menu": False,
            "description": "Base Node in Infrahub.",
            "label": "Node",
        },
        {
            "name": "Owner",
            "namespace": "Lineage",
            "description": "Any Entities that is responsible for some data.",
            "label": "Owner",
            "include_in_menu": False,
            "documentation": "/topics/metadata",
        },
        core_profile_schema_definition,
        {
            "name": "Source",
            "namespace": "Lineage",
            "description": "Any Entities that stores or produces data.",
            "label": "Source",
            "include_in_menu": False,
            "documentation": "/topics/metadata",
        },
        {
            "name": "Comment",
            "namespace": "Core",
            "description": "A comment on a Proposed Change",
            "label": "Comment",
            "display_labels": ["text__value"],
            "order_by": ["created_at__value"],
            "include_in_menu": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "text", "kind": "TextArea", "unique": False, "optional": False},
                {"name": "created_at", "kind": "DateTime", "optional": True},
            ],
            "relationships": [
                {
                    "name": "created_by",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "cardinality": "one",
                    "identifier": "comment__account",
                },
            ],
        },
        {
            "name": "Thread",
            "namespace": "Core",
            "description": "A thread on a Proposed Change",
            "label": "Thread",
            "order_by": ["created_at__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "include_in_menu": False,
            "attributes": [
                {"name": "label", "kind": "Text", "optional": True},
                {"name": "resolved", "kind": "Boolean", "default_value": False},
                {"name": "created_at", "kind": "DateTime", "optional": True},
            ],
            "relationships": [
                {
                    "name": "change",
                    "peer": InfrahubKind.PROPOSEDCHANGE,
                    "identifier": "proposedchange__thread",
                    "kind": "Parent",
                    "optional": False,
                    "cardinality": "one",
                },
                {
                    "name": "comments",
                    "peer": InfrahubKind.THREADCOMMENT,
                    "identifier": "thread__threadcomment",
                    "kind": "Component",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                },
                {
                    "name": "created_by",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "identifier": "thread__account",
                    "optional": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "cardinality": "one",
                },
            ],
        },
        {
            "name": "Group",
            "namespace": "Core",
            "description": "Generic Group Object.",
            "label": "Group",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["label__value"],
            "include_in_menu": False,
            "icon": "mdi:group",
            "hierarchical": True,
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "label", "kind": "Text", "optional": True},
                {"name": "description", "kind": "Text", "optional": True},
                {
                    "name": "group_type",
                    "kind": "Text",
                    "enum": ["default", "internal"],
                    "default_value": "default",
                    "optional": False,
                },
            ],
            "relationships": [
                {
                    "name": "members",
                    "peer": InfrahubKind.NODE,
                    "optional": True,
                    "identifier": "group_member",
                    "cardinality": "many",
                    "branch": BranchSupportType.AWARE,
                },
                {
                    "name": "subscribers",
                    "peer": InfrahubKind.NODE,
                    "optional": True,
                    "identifier": "group_subscriber",
                    "cardinality": "many",
                },
            ],
        },
        {
            "name": "Validator",
            "namespace": "Core",
            "description": "",
            "include_in_menu": False,
            "label": "Validator",
            "order_by": ["started_at__value"],
            "display_labels": ["label__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "label", "kind": "Text", "optional": True},
                {
                    "name": "state",
                    "kind": "Text",
                    "enum": ValidatorState.available_types(),
                    "default_value": ValidatorState.QUEUED.value,
                },
                {
                    "name": "conclusion",
                    "kind": "Text",
                    "enum": ValidatorConclusion.available_types(),
                    "default_value": ValidatorConclusion.UNKNOWN.value,
                },
                {"name": "completed_at", "kind": "DateTime", "optional": True},
                {"name": "started_at", "kind": "DateTime", "optional": True},
            ],
            "relationships": [
                {
                    "name": "proposed_change",
                    "peer": InfrahubKind.PROPOSEDCHANGE,
                    "kind": "Parent",
                    "optional": False,
                    "cardinality": "one",
                    "identifier": "proposed_change__validator",
                },
                {
                    "name": "checks",
                    "peer": "CoreCheck",
                    "kind": "Component",
                    "optional": True,
                    "cardinality": "many",
                    "identifier": "validator__check",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                },
            ],
        },
        {
            "name": "Check",
            "namespace": "Core",
            "description": "",
            "display_labels": ["label__value"],
            "include_in_menu": False,
            "label": "Check",
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "name", "kind": "Text", "optional": True},
                {"name": "label", "kind": "Text", "optional": True},
                {"name": "origin", "kind": "Text", "optional": False},
                {
                    "name": "kind",
                    "kind": "Text",
                    "regex": "^[A-Z][a-zA-Z0-9]+$",
                    "optional": False,
                    "min_length": DEFAULT_KIND_MIN_LENGTH,
                    "max_length": DEFAULT_KIND_MAX_LENGTH,
                },
                {"name": "message", "kind": "TextArea", "optional": True},
                {
                    "name": "conclusion",
                    "kind": "Text",
                    "enum": ValidatorConclusion.available_types(),
                    "default_value": ValidatorConclusion.UNKNOWN.value,
                    "optional": True,
                },
                {
                    "name": "severity",
                    "kind": "Text",
                    "enum": Severity.available_types(),
                    "default_value": Severity.INFO.value,
                    "optional": True,
                },
                {"name": "created_at", "kind": "DateTime", "optional": True},
            ],
            "relationships": [
                {
                    "name": "validator",
                    "peer": InfrahubKind.VALIDATOR,
                    "identifier": "validator__check",
                    "kind": "Parent",
                    "optional": False,
                    "cardinality": "one",
                },
            ],
        },
        {
            "name": "Transformation",
            "namespace": "Core",
            "description": "Generic Transformation Object.",
            "include_in_menu": False,
            "icon": "mdi:cog-transfer",
            "label": "Transformation",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["label__value"],
            "branch": BranchSupportType.AWARE.value,
            "documentation": "/topics/proposed-change",
            "uniqueness_constraints": [["name__value"]],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "label", "kind": "Text", "optional": True},
                {"name": "description", "kind": "Text", "optional": True},
                {"name": "timeout", "kind": "Number", "default_value": 10},
            ],
            "relationships": [
                {
                    "name": "query",
                    "peer": InfrahubKind.GRAPHQLQUERY,
                    "identifier": "graphql_query__transformation",
                    "kind": "Attribute",
                    "cardinality": "one",
                    "optional": False,
                },
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "cardinality": "one",
                    "identifier": "repository__transformation",
                    "optional": False,
                },
                {
                    "name": "tags",
                    "peer": InfrahubKind.TAG,
                    "kind": "Attribute",
                    "optional": True,
                    "cardinality": "many",
                },
            ],
        },
        {
            "name": "ArtifactTarget",
            "include_in_menu": False,
            "namespace": "Core",
            "description": "Extend a node to be associated with artifacts",
            "label": "Artifact Target",
            "relationships": [
                {
                    "name": "artifacts",
                    "peer": InfrahubKind.ARTIFACT,
                    "optional": True,
                    "cardinality": "many",
                    "kind": "Generic",
                    "identifier": "artifact__node",
                },
            ],
        },
        {
            "name": "TaskTarget",
            "include_in_menu": False,
            "namespace": "Core",
            "description": "Extend a node to be associated with tasks",
            "label": "Task Target",
        },
        {
            "name": "Webhook",
            "namespace": "Core",
            "description": "A webhook that connects to an external integration",
            "label": "Webhook",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "include_in_menu": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "uniqueness_constraints": [["name__value"]],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True, "order_weight": 1000},
                {"name": "description", "kind": "Text", "optional": True, "order_weight": 2000},
                {"name": "url", "kind": "URL", "order_weight": 3000},
                {
                    "name": "validate_certificates",
                    "kind": "Boolean",
                    "default_value": True,
                    "optional": True,
                    "order_weight": 5000,
                },
            ],
        },
        {
            "name": "GenericRepository",
            "namespace": "Core",
            "label": "Git Repository",
            "description": "A Git Repository integrated with Infrahub",
            "include_in_menu": False,
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "icon": "mdi:source-repository",
            "branch": BranchSupportType.AGNOSTIC.value,
            "uniqueness_constraints": [["name__value"], ["location__value"]],
            "documentation": "/topics/repository",
            "attributes": [
                {
                    "name": "name",
                    "regex": "^[^/]*$",
                    "kind": "Text",
                    "unique": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "order_weight": 1000,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "description",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "order_weight": 2000,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "location",
                    "kind": "Text",
                    "unique": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "order_weight": 3000,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "internal_status",
                    "kind": "Dropdown",
                    "choices": [
                        {
                            "name": RepositoryInternalStatus.STAGING.value,
                            "label": "Staging",
                            "description": "Repository was recently added to this branch.",
                            "color": "#fef08a",
                        },
                        {
                            "name": RepositoryInternalStatus.ACTIVE.value,
                            "label": "Active",
                            "description": "Repository is actively being synced for this branch",
                            "color": "#86efac",
                        },
                        {
                            "name": RepositoryInternalStatus.INACTIVE.value,
                            "label": "Inactive",
                            "description": "Repository is not active on this branch.",
                            "color": "#e5e7eb",
                        },
                    ],
                    "default_value": "inactive",
                    "optional": False,
                    "branch": BranchSupportType.LOCAL.value,
                    "order_weight": 7000,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "operational_status",
                    "kind": "Dropdown",
                    "choices": [
                        {
                            "name": RepositoryOperationalStatus.UNKNOWN.value,
                            "label": "Unknown",
                            "description": "Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                            "color": "#9ca3af",
                        },
                        {
                            "name": RepositoryOperationalStatus.ONLINE.value,
                            "label": "Online",
                            "description": "Repository connection is working",
                            "color": "#86efac",
                        },
                        {
                            "name": RepositoryOperationalStatus.ERROR_CRED.value,
                            "label": "Credential Error",
                            "description": "Repository can't be synced due to some credential error(s)",
                            "color": "#f87171",
                        },
                        {
                            "name": RepositoryOperationalStatus.ERROR_CONNECTION.value,
                            "label": "Connectivity Error",
                            "description": "Repository can't be synced due to some connectivity error(s)",
                            "color": "#f87171",
                        },
                        {
                            "name": RepositoryOperationalStatus.ERROR.value,
                            "label": "Error",
                            "description": "Repository can't be synced due to an unknown error",
                            "color": "#ef4444",
                        },
                    ],
                    "optional": False,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "default_value": RepositoryOperationalStatus.UNKNOWN.value,
                    "order_weight": 5000,
                },
                {
                    "name": "sync_status",
                    "kind": "Dropdown",
                    "choices": [
                        {
                            "name": RepositorySyncStatus.UNKNOWN.value,
                            "label": "Unknown",
                            "description": "Status of the repository is unknown and mostlikely because it hasn't been synced yet",
                            "color": "#9ca3af",
                        },
                        {
                            "name": RepositorySyncStatus.ERROR_IMPORT.value,
                            "label": "Import Error",
                            "description": "Repository import error observed",
                            "color": "#f87171",
                        },
                        {
                            "name": RepositorySyncStatus.IN_SYNC.value,
                            "label": "In Sync",
                            "description": "The repository is syncing correctly",
                            "color": "#60a5fa",
                        },
                        {
                            "name": RepositorySyncStatus.SYNCING.value,
                            "label": "Syncing",
                            "description": "A sync job is currently running against the repository.",
                            "color": "#a855f7",
                        },
                    ],
                    "optional": False,
                    "branch": BranchSupportType.LOCAL.value,
                    "default_value": RepositorySyncStatus.UNKNOWN.value,
                    "order_weight": 6000,
                },
            ],
            "relationships": [
                {
                    "name": "credential",
                    "peer": InfrahubKind.CREDENTIAL,
                    "identifier": "gitrepository__credential",
                    "kind": "Attribute",
                    "optional": True,
                    "cardinality": "one",
                    "order_weight": 4000,
                },
                {
                    "name": "tags",
                    "peer": InfrahubKind.TAG,
                    "kind": "Attribute",
                    "optional": True,
                    "cardinality": "many",
                    "order_weight": 8000,
                },
                {
                    "name": "transformations",
                    "peer": InfrahubKind.TRANSFORM,
                    "identifier": "repository__transformation",
                    "optional": True,
                    "cardinality": "many",
                    "order_weight": 10000,
                },
                {
                    "name": "queries",
                    "peer": InfrahubKind.GRAPHQLQUERY,
                    "identifier": "graphql_query__repository",
                    "optional": True,
                    "cardinality": "many",
                    "order_weight": 9000,
                },
                {
                    "name": "checks",
                    "peer": InfrahubKind.CHECKDEFINITION,
                    "identifier": "check_definition__repository",
                    "optional": True,
                    "cardinality": "many",
                    "order_weight": 11000,
                },
                {
                    "name": "generators",
                    "peer": InfrahubKind.GENERATORDEFINITION,
                    "identifier": "generator_definition__repository",
                    "optional": True,
                    "cardinality": "many",
                    "order_weight": 12000,
                },
            ],
        },
        {
            "name": "IPNamespace",
            "namespace": "Builtin",
            "label": "IP Namespace",
            "description": "A generic container for IP prefixes and IP addresses",
            "include_in_menu": False,
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "icon": "mdi:format-list-group",
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "generate_profile": False,
            "attributes": [
                {
                    "name": "name",
                    "kind": "Text",
                    "unique": True,
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 1000,
                },
                {
                    "name": "description",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 2000,
                },
            ],
            "relationships": [
                {
                    "name": "ip_prefixes",
                    "label": "IP Prefixes",
                    "peer": InfrahubKind.IPPREFIX,
                    "identifier": "ip_namespace__ip_prefix",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "ip_addresses",
                    "label": "IP Addresses",
                    "peer": InfrahubKind.IPADDRESS,
                    "identifier": "ip_namespace__ip_address",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                    "allow_override": AllowOverrideType.NONE,
                },
            ],
        },
        {
            "name": "IPPrefix",
            "label": "IP Prefix",
            "namespace": "Builtin",
            "description": "IPv6 or IPv4 prefix also referred as network",
            "include_in_menu": False,
            "default_filter": "prefix__value",
            "order_by": ["prefix__version", "prefix__binary_address", "prefix__prefixlen"],
            "display_labels": ["prefix__value"],
            "icon": "mdi:ip-network",
            "branch": BranchSupportType.AWARE.value,
            "hierarchical": True,
            "attributes": [
                {
                    "name": "prefix",
                    "kind": "IPNetwork",
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 1000,
                },
                {
                    "name": "description",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 2000,
                },
                {
                    "name": "member_type",
                    "kind": "Dropdown",
                    "choices": [
                        {
                            "name": "prefix",
                            "label": "Prefix",
                            "description": "Prefix serves as container for other prefixes",
                        },
                        {
                            "name": "address",
                            "label": "Address",
                            "description": "Prefix serves as subnet for IP addresses",
                        },
                    ],
                    "branch": BranchSupportType.AWARE.value,
                    "default_value": "address",
                    "order_weight": 3000,
                },
                {
                    "name": "is_pool",
                    "kind": "Boolean",
                    "branch": BranchSupportType.AWARE.value,
                    "default_value": False,
                    "order_weight": 4000,
                    "description": "All IP addresses within this prefix are considered usable",
                },
                {
                    "name": "is_top_level",
                    "kind": "Boolean",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "utilization",
                    "kind": "Number",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "netmask",
                    "kind": "Text",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "hostmask",
                    "kind": "Text",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "network_address",
                    "kind": "Text",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "broadcast_address",
                    "kind": "Text",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
            ],
            "relationships": [
                {
                    "name": "ip_namespace",
                    "label": "IP Namespace",
                    "peer": InfrahubKind.IPNAMESPACE,
                    "identifier": "ip_namespace__ip_prefix",
                    "optional": True,
                    "cardinality": "one",
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "ip_addresses",
                    "label": "IP Addresses",
                    "peer": InfrahubKind.IPADDRESS,
                    "identifier": "ip_prefix__ip_address",
                    "optional": True,
                    "cardinality": "many",
                    "allow_override": AllowOverrideType.NONE,
                    "read_only": True,
                },
                {
                    "name": "resource_pool",
                    "peer": "CoreIPAddressPool",
                    "identifier": "ipaddresspool__resource",
                    "cardinality": "many",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "optional": True,
                    "read_only": True,
                },
            ],
        },
        {
            "name": "IPAddress",
            "label": "IP Address",
            "namespace": "Builtin",
            "description": "IPv6 or IPv4 address",
            "include_in_menu": False,
            "default_filter": "address__value",
            "order_by": ["address__version", "address__binary_address"],
            "display_labels": ["address__value"],
            "icon": "mdi:ip-outline",
            "branch": BranchSupportType.AWARE.value,
            "attributes": [
                {
                    "name": "address",
                    "kind": "IPHost",
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 1000,
                },
                {
                    "name": "description",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 2000,
                },
            ],
            "relationships": [
                {
                    "name": "ip_namespace",
                    "label": "IP Namespace",
                    "peer": InfrahubKind.IPNAMESPACE,
                    "identifier": "ip_namespace__ip_address",
                    "optional": True,
                    "cardinality": "one",
                    "allow_override": AllowOverrideType.NONE,
                },
                {
                    "name": "ip_prefix",
                    "label": "IP Prefix",
                    "peer": InfrahubKind.IPPREFIX,
                    "identifier": "ip_prefix__ip_address",
                    "optional": True,
                    "cardinality": "one",
                    "allow_override": AllowOverrideType.NONE,
                    "read_only": True,
                },
            ],
        },
        {
            "name": "ResourcePool",
            "namespace": "Core",
            "label": "Resource Pool",
            "description": "The resource manager contains pools of resources to allow for automatic assignments.",
            "include_in_menu": False,
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "human_friendly_id": ["name__value"],
            "icon": "mdi:view-grid-outline",
            "branch": BranchSupportType.AGNOSTIC.value,
            "uniqueness_constraints": [["name__value"]],
            "generate_profile": False,
            "attributes": [
                {
                    "name": "name",
                    "kind": "Text",
                    "order_weight": 1000,
                    "unique": True,
                },
                {
                    "name": "description",
                    "kind": "Text",
                    "optional": True,
                    "order_weight": 2000,
                },
            ],
        },
        {
            "name": "GenericAccount",
            "namespace": "Core",
            "description": "User Account for Infrahub",
            "include_in_menu": False,
            "label": "Account",
            "icon": "mdi:account",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["label__value"],
            "human_friendly_id": ["name__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "documentation": "/topics/auth",
            "uniqueness_constraints": [["name__value"]],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "password", "kind": "HashedPassword", "unique": False},
                {"name": "label", "kind": "Text", "optional": True},
                {"name": "description", "kind": "Text", "optional": True},
                {
                    "name": "account_type",
                    "kind": "Text",
                    "default_value": AccountType.USER.value,
                    "enum": AccountType.available_types(),
                },
                {
                    "name": "status",
                    "kind": "Dropdown",
                    "choices": [
                        {
                            "name": AccountStatus.ACTIVE.value,
                            "label": "Active",
                            "description": "Account is allowed to login",
                            "color": "#52be80",
                        },
                        {
                            "name": AccountStatus.INACTIVE.value,
                            "label": "Inactive",
                            "description": "Account is not allowed to login",
                            "color": "#e74c3c",
                        },
                    ],
                    "default_value": AccountStatus.ACTIVE.value,
                },
            ],
            "relationships": [
                {"name": "tokens", "peer": InfrahubKind.ACCOUNTTOKEN, "optional": True, "cardinality": "many"}
            ],
        },
        {
            "name": "BasePermission",
            "namespace": "Core",
            "description": "A permission grants right to an account",
            "label": "Base permission",
            "icon": "mdi:user-key",
            "include_in_menu": False,
            "generate_profile": False,
            "attributes": [
                {"name": "description", "kind": "Text", "optional": True},
                {
                    "name": "identifier",
                    "kind": "Text",
                    "read_only": True,
                    "optional": True,
                    "allow_override": AllowOverrideType.NONE,
                },
            ],
            "relationships": [
                {
                    "name": "roles",
                    "peer": InfrahubKind.ACCOUNTROLE,
                    "optional": True,
                    "identifier": "role__permissions",
                    "cardinality": "many",
                    "kind": "Attribute",
                }
            ],
        },
        {
            "name": "Credential",
            "namespace": "Core",
            "description": "A credential that could be referenced to access external services.",
            "include_in_menu": False,
            "label": "Credential",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["label__value"],
            "icon": "mdi:key-variant",
            "human_friendly_id": ["name__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "uniqueness_constraints": [["name__value"]],
            "documentation": "/topics/auth",
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True, "order_weight": 1000},
                {"name": "label", "kind": "Text", "optional": True, "order_weight": 2000},
                {"name": "description", "kind": "Text", "optional": True, "order_weight": 3000},
            ],
        },
        generic_menu_item,
    ],
    "nodes": [
        menu_item,
        {
            "name": "StandardGroup",
            "namespace": "Core",
            "description": "Group of nodes of any kind.",
            "include_in_menu": False,
            "icon": "mdi:account-group",
            "label": "Standard Group",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.AWARE.value,
            "inherit_from": [InfrahubKind.GENERICGROUP],
            "generate_profile": False,
        },
        {
            "name": "GeneratorGroup",
            "namespace": "Core",
            "description": "Group of nodes that are created by a generator.",
            "include_in_menu": False,
            "icon": "mdi:state-machine",
            "label": "Generator Group",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.LOCAL.value,
            "inherit_from": [InfrahubKind.GENERICGROUP],
            "generate_profile": False,
        },
        {
            "name": "GraphQLQueryGroup",
            "namespace": "Core",
            "description": "Group of nodes associated with a given GraphQLQuery.",
            "include_in_menu": False,
            "icon": "mdi:account-group",
            "label": "GraphQL Query Group",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.LOCAL.value,
            "inherit_from": [InfrahubKind.GENERICGROUP],
            "generate_profile": False,
            "attributes": [
                {"name": "parameters", "kind": "JSON", "optional": True},
            ],
            "relationships": [
                {
                    "name": "query",
                    "peer": InfrahubKind.GRAPHQLQUERY,
                    "optional": False,
                    "cardinality": "one",
                    "kind": "Attribute",
                },
            ],
        },
        {
            "name": "Tag",
            "namespace": "Builtin",
            "description": "Standard Tag object to attached to other objects to provide some context.",
            "include_in_menu": True,
            "icon": "mdi:tag-multiple",
            "label": "Tag",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "description", "kind": "Text", "optional": True},
            ],
        },
        {
            "name": "Account",
            "namespace": "Core",
            "description": "User Account for Infrahub",
            "include_in_menu": False,
            "label": "Account",
            "icon": "mdi:account",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["label__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICACCOUNT],
        },
        {
            "name": "AccountToken",
            "namespace": "Internal",
            "description": "Token for User Account",
            "include_in_menu": False,
            "label": "Account Token",
            "default_filter": "token__value",
            "display_labels": ["token__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "uniqueness_constraints": [["token__value"]],
            "documentation": "/topics/auth",
            "attributes": [
                {"name": "name", "kind": "Text", "optional": True},
                {"name": "token", "kind": "Text", "unique": True},
                {"name": "expiration", "kind": "DateTime", "optional": True},
            ],
            "relationships": [
                {
                    "name": "account",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": False,
                    "cardinality": "one",
                    "identifier": "account__token",
                },
            ],
        },
        {
            "name": "PasswordCredential",
            "namespace": "Core",
            "description": "Username/Password based credential",
            "include_in_menu": False,
            "label": "Username / Password",
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.CREDENTIAL],
            "attributes": [
                {
                    "name": "username",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "order_weight": 6000,
                },
                {
                    "name": "password",
                    "kind": "Password",
                    "optional": True,
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "order_weight": 7000,
                },
            ],
        },
        {
            "name": "RefreshToken",
            "namespace": "Internal",
            "description": "Refresh Token",
            "include_in_menu": False,
            "label": "Refresh Token",
            "display_labels": [],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "expiration", "kind": "DateTime", "optional": False},
            ],
            "relationships": [
                {
                    "name": "account",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": False,
                    "cardinality": "one",
                    "identifier": "account__refreshtoken",
                },
            ],
        },
        {
            "name": "ProposedChange",
            "namespace": "Core",
            "description": "Metadata related to a proposed change",
            "include_in_menu": False,
            "icon": "mdi:file-replace-outline",
            "label": "Proposed Change",
            "default_filter": "name__value",
            "display_labels": ["name__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.TASKTARGET],
            "documentation": "/topics/proposed-change",
            "attributes": [
                {"name": "name", "kind": "Text", "optional": False},
                {"name": "description", "kind": "TextArea", "optional": True},
                {"name": "source_branch", "kind": "Text", "optional": False},
                {"name": "destination_branch", "kind": "Text", "optional": False},
                {
                    "name": "state",
                    "kind": "Text",
                    "enum": ProposedChangeState.available_types(),
                    "default_value": ProposedChangeState.OPEN.value,
                    "optional": True,
                },
            ],
            "relationships": [
                {
                    "name": "approved_by",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": True,
                    "cardinality": "many",
                    "kind": "Attribute",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "identifier": "coreaccount__proposedchange_approved_by",
                },
                {
                    "name": "reviewers",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": True,
                    "kind": "Attribute",
                    "cardinality": "many",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "identifier": "coreaccount__proposedchange_reviewed_by",
                },
                {
                    "name": "created_by",
                    "peer": InfrahubKind.GENERICACCOUNT,
                    "optional": True,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "identifier": "coreaccount__proposedchange_created_by",
                },
                {
                    "name": "comments",
                    "peer": InfrahubKind.CHANGECOMMENT,
                    "kind": "Component",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                },
                {
                    "name": "threads",
                    "peer": InfrahubKind.THREAD,
                    "identifier": "proposedchange__thread",
                    "kind": "Component",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                },
                {
                    "name": "validations",
                    "peer": InfrahubKind.VALIDATOR,
                    "kind": "Component",
                    "identifier": "proposed_change__validator",
                    "optional": True,
                    "cardinality": "many",
                    "on_delete": RelationshipDeleteBehavior.CASCADE,
                },
            ],
        },
        {
            "name": "ChangeThread",
            "namespace": "Core",
            "description": "A thread on proposed change",
            "include_in_menu": False,
            "label": "Change Thread",
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.THREAD],
            "generate_profile": False,
            "attributes": [],
            "relationships": [],
        },
        {
            "name": "FileThread",
            "namespace": "Core",
            "description": "A thread related to a file on a proposed change",
            "include_in_menu": False,
            "label": "Thread - File",
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.THREAD],
            "generate_profile": False,
            "attributes": [
                {"name": "file", "kind": "Text", "optional": True},
                {"name": "commit", "kind": "Text", "optional": True},
                {"name": "line_number", "kind": "Number", "optional": True},
            ],
            "relationships": [
                {
                    "name": "repository",
                    "peer": InfrahubKind.REPOSITORY,
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
            ],
        },
        {
            "name": "ArtifactThread",
            "namespace": "Core",
            "description": "A thread related to an artifact on a proposed change",
            "include_in_menu": False,
            "label": "Thread - Artifact",
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.THREAD],
            "generate_profile": False,
            "attributes": [
                {"name": "artifact_id", "kind": "Text", "optional": True},
                {"name": "storage_id", "kind": "Text", "optional": True},
                {"name": "line_number", "kind": "Number", "optional": True},
            ],
            "relationships": [],
        },
        {
            "name": "ObjectThread",
            "namespace": "Core",
            "description": "A thread related to an object on a proposed change",
            "include_in_menu": False,
            "label": "Thread - Object",
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.THREAD],
            "generate_profile": False,
            "attributes": [
                {"name": "object_path", "kind": "Text", "optional": False},
            ],
            "relationships": [],
        },
        {
            "name": "ChangeComment",
            "namespace": "Core",
            "description": "A comment on proposed change",
            "include_in_menu": False,
            "label": "Change Comment",
            "default_filter": "text__value",
            "display_labels": ["text__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.COMMENT],
            "generate_profile": False,
            "relationships": [
                {
                    "name": "change",
                    "kind": "Parent",
                    "peer": InfrahubKind.PROPOSEDCHANGE,
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "ThreadComment",
            "namespace": "Core",
            "description": "A comment on thread within a Proposed Change",
            "include_in_menu": False,
            "label": "Thread Comment",
            "default_filter": "text__value",
            "display_labels": ["text__value"],
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [InfrahubKind.COMMENT],
            "generate_profile": False,
            "attributes": [],
            "relationships": [
                {
                    "name": "thread",
                    "peer": InfrahubKind.THREAD,
                    "kind": "Parent",
                    "identifier": "thread__threadcomment",
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "Repository",
            "namespace": "Core",
            "description": "A Git Repository integrated with Infrahub",
            "include_in_menu": False,
            "icon": "mdi:source-repository",
            "label": "Repository",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [
                InfrahubKind.LINEAGEOWNER,
                InfrahubKind.LINEAGESOURCE,
                InfrahubKind.GENERICREPOSITORY,
                InfrahubKind.TASKTARGET,
            ],
            "documentation": "/topics/repository",
            "attributes": [
                {"name": "default_branch", "kind": "Text", "default_value": "main", "order_weight": 6000},
                {
                    "name": "commit",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.LOCAL.value,
                    "order_weight": 7000,
                },
            ],
        },
        {
            "name": "ReadOnlyRepository",
            "namespace": "Core",
            "description": "A Git Repository integrated with Infrahub, Git-side will not be updated",
            "include_in_menu": False,
            "label": "Read-Only Repository",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "inherit_from": [
                InfrahubKind.LINEAGEOWNER,
                InfrahubKind.LINEAGESOURCE,
                InfrahubKind.GENERICREPOSITORY,
                InfrahubKind.TASKTARGET,
            ],
            "documentation": "/topics/repository",
            "attributes": [
                {
                    "name": "ref",
                    "kind": "Text",
                    "default_value": "main",
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 6000,
                },
                {
                    "name": "commit",
                    "kind": "Text",
                    "optional": True,
                    "branch": BranchSupportType.AWARE.value,
                    "order_weight": 7000,
                },
            ],
        },
        {
            "name": "TransformJinja2",
            "namespace": "Core",
            "description": "A file rendered from a Jinja2 template",
            "include_in_menu": False,
            "label": "Transform Jinja2",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "inherit_from": [InfrahubKind.TRANSFORM],
            "generate_profile": False,
            "branch": BranchSupportType.AWARE.value,
            "documentation": "/topics/transformation",
            "attributes": [
                {"name": "template_path", "kind": "Text"},
            ],
        },
        {
            "name": "DataCheck",
            "namespace": "Core",
            "description": "A check related to some Data",
            "include_in_menu": False,
            "label": "Data Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "conflicts", "kind": "JSON"},
                {"name": "keep_branch", "enum": BranchConflictKeep.available_types(), "kind": "Text", "optional": True},
                {"name": "enriched_conflict_id", "kind": "Text", "optional": True},
            ],
        },
        {
            "name": "StandardCheck",
            "namespace": "Core",
            "description": "A standard check",
            "include_in_menu": False,
            "label": "Standard Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
        },
        {
            "name": "SchemaCheck",
            "namespace": "Core",
            "description": "A check related to the schema",
            "include_in_menu": False,
            "label": "Schema Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "conflicts", "kind": "JSON"},
                {"name": "enriched_conflict_id", "kind": "Text", "optional": True},
            ],
        },
        {
            "name": "FileCheck",
            "namespace": "Core",
            "description": "A check related to a file in a Git Repository",
            "include_in_menu": False,
            "label": "File Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "files", "kind": "List", "optional": True},
                {"name": "commit", "kind": "Text", "optional": True},
            ],
        },
        {
            "name": "ArtifactCheck",
            "namespace": "Core",
            "description": "A check related to an artifact",
            "include_in_menu": False,
            "label": "Artifact Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {"name": "changed", "kind": "Boolean", "optional": True},
                {"name": "checksum", "kind": "Text", "optional": True},
                {"name": "artifact_id", "kind": "Text", "optional": True},
                {"name": "storage_id", "kind": "Text", "optional": True},
                {"name": "line_number", "kind": "Number", "optional": True},
            ],
        },
        {
            "name": "GeneratorCheck",
            "namespace": "Core",
            "description": "A check related to a Generator instance",
            "include_in_menu": False,
            "label": "Generator Check",
            "display_labels": ["label__value"],
            "inherit_from": ["CoreCheck"],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {
                    "name": "instance",
                    "kind": "Text",
                    "optional": False,
                },
            ],
        },
        {
            "name": "DataValidator",
            "namespace": "Core",
            "description": "A check to validate the data integrity between two branches",
            "include_in_menu": False,
            "label": "Data Validator",
            "display_labels": ["label__value"],
            "inherit_from": [InfrahubKind.VALIDATOR],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
        },
        {
            "name": "RepositoryValidator",
            "namespace": "Core",
            "description": "A Validator related to a specific repository",
            "include_in_menu": False,
            "label": "Repository Validator",
            "display_labels": ["label__value"],
            "inherit_from": [InfrahubKind.VALIDATOR],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "relationships": [
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
            ],
        },
        {
            "name": "UserValidator",
            "namespace": "Core",
            "description": "A Validator related to a user defined checks in a repository",
            "include_in_menu": False,
            "label": "User Validator",
            "display_labels": ["label__value"],
            "inherit_from": [InfrahubKind.VALIDATOR],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "relationships": [
                {
                    "name": "check_definition",
                    "peer": InfrahubKind.CHECKDEFINITION,
                    "kind": "Attribute",
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
            ],
        },
        {
            "name": "SchemaValidator",
            "namespace": "Core",
            "description": "A validator related to the schema",
            "include_in_menu": False,
            "label": "Schema Validator",
            "display_labels": ["label__value"],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.VALIDATOR],
            "branch": BranchSupportType.AGNOSTIC.value,
        },
        {
            "name": "ArtifactValidator",
            "namespace": "Core",
            "description": "A validator related to the artifacts",
            "include_in_menu": False,
            "label": "Artifact Validator",
            "display_labels": ["label__value"],
            "inherit_from": [InfrahubKind.VALIDATOR],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "relationships": [
                {
                    "name": "definition",
                    "peer": InfrahubKind.ARTIFACTDEFINITION,
                    "kind": "Attribute",
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
            ],
        },
        {
            "name": "GeneratorValidator",
            "namespace": "Core",
            "description": "A validator related to generators",
            "include_in_menu": False,
            "label": "Generator Validator",
            "display_labels": ["label__value"],
            "inherit_from": [InfrahubKind.VALIDATOR],
            "generate_profile": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "relationships": [
                {
                    "name": "definition",
                    "peer": InfrahubKind.GENERATORDEFINITION,
                    "kind": "Attribute",
                    "optional": False,
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                },
            ],
        },
        {
            "name": "CheckDefinition",
            "namespace": "Core",
            "include_in_menu": False,
            "icon": "mdi:check-all",
            "label": "Check Definition",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.TASKTARGET],
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "description", "kind": "Text", "optional": True},
                {"name": "file_path", "kind": "Text"},
                {"name": "class_name", "kind": "Text"},
                {"name": "timeout", "kind": "Number", "default_value": 10},
                {"name": "parameters", "kind": "JSON", "optional": True},
            ],
            "relationships": [
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "cardinality": "one",
                    "identifier": "check_definition__repository",
                    "optional": False,
                },
                {
                    "name": "query",
                    "peer": InfrahubKind.GRAPHQLQUERY,
                    "kind": "Attribute",
                    "identifier": "check_definition__graphql_query",
                    "cardinality": "one",
                    "optional": True,
                },
                {
                    "name": "targets",
                    "peer": InfrahubKind.GENERICGROUP,
                    "kind": "Attribute",
                    "identifier": "check_definition___group",
                    "cardinality": "one",
                    "optional": True,
                },
                {
                    "name": "tags",
                    "peer": InfrahubKind.TAG,
                    "kind": "Attribute",
                    "optional": True,
                    "cardinality": "many",
                },
            ],
        },
        {
            "name": "TransformPython",
            "namespace": "Core",
            "description": "A transform function written in Python",
            "include_in_menu": False,
            "label": "Transform Python",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "inherit_from": [InfrahubKind.TRANSFORM],
            "generate_profile": False,
            "branch": BranchSupportType.AWARE.value,
            "documentation": "/topics/transformation",
            "attributes": [
                {"name": "file_path", "kind": "Text"},
                {"name": "class_name", "kind": "Text"},
            ],
        },
        {
            "name": "GraphQLQuery",
            "namespace": "Core",
            "description": "A pre-defined GraphQL Query",
            "include_in_menu": False,
            "icon": "mdi:graphql",
            "label": "GraphQL Query",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "generate_profile": False,
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "documentation": "/topics/graphql",
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "description", "kind": "Text", "optional": True},
                {"name": "query", "kind": "TextArea"},
                {
                    "name": "variables",
                    "kind": "JSON",
                    "description": "variables in use in the query",
                    "optional": True,
                    "read_only": True,
                },
                {
                    "name": "operations",
                    "kind": "List",
                    "description": "Operations in use in the query, valid operations: 'query', 'mutation' or 'subscription'",
                    "read_only": True,
                    "optional": True,
                },
                {
                    "name": "models",
                    "kind": "List",
                    "description": "List of models associated with this query",
                    "read_only": True,
                    "optional": True,
                },
                {
                    "name": "depth",
                    "kind": "Number",
                    "description": "number of nested levels in the query",
                    "read_only": True,
                    "optional": True,
                },
                {
                    "name": "height",
                    "kind": "Number",
                    "description": "total number of fields requested in the query",
                    "read_only": True,
                    "optional": True,
                },
            ],
            "relationships": [
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "identifier": "graphql_query__repository",
                    "cardinality": "one",
                    "optional": True,
                },
                {
                    "name": "tags",
                    "peer": InfrahubKind.TAG,
                    "kind": "Attribute",
                    "optional": True,
                    "cardinality": "many",
                },
            ],
        },
        {
            "name": "Artifact",
            "namespace": "Core",
            "label": "Artifact",
            "include_in_menu": False,
            "icon": "mdi:file-document-outline",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.LOCAL.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.TASKTARGET],
            "documentation": "/topics/artifact",
            "attributes": [
                {"name": "name", "kind": "Text"},
                {
                    "name": "status",
                    "kind": "Text",
                    "enum": ArtifactStatus.available_types(),
                },
                {
                    "name": "content_type",
                    "kind": "Text",
                    "enum": ContentType.available_types(),
                },
                {
                    "name": "checksum",
                    "kind": "Text",
                    "optional": True,
                },
                {
                    "name": "storage_id",
                    "kind": "Text",
                    "optional": True,
                    "description": "ID of the file in the object store",
                },
                {"name": "parameters", "kind": "JSON", "optional": True},
            ],
            "relationships": [
                {
                    "name": "object",
                    "peer": InfrahubKind.ARTIFACTTARGET,
                    "kind": "Attribute",
                    "identifier": "artifact__node",
                    "cardinality": "one",
                    "optional": False,
                },
                {
                    "name": "definition",
                    "peer": InfrahubKind.ARTIFACTDEFINITION,
                    "kind": "Attribute",
                    "identifier": "artifact__artifact_definition",
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "ArtifactDefinition",
            "namespace": "Core",
            "include_in_menu": False,
            "icon": "mdi:file-document-multiple-outline",
            "label": "Artifact Definition",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.AWARE.value,
            "generate_profile": False,
            "uniqueness_constraints": [["name__value"]],
            "inherit_from": [InfrahubKind.TASKTARGET],
            "documentation": "/topics/artifact",
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "artifact_name", "kind": "Text"},
                {"name": "description", "kind": "Text", "optional": True},
                {"name": "parameters", "kind": "JSON"},
                {
                    "name": "content_type",
                    "kind": "Text",
                    "enum": ContentType.available_types(),
                },
            ],
            "relationships": [
                {
                    "name": "targets",
                    "peer": InfrahubKind.GENERICGROUP,
                    "kind": "Attribute",
                    "identifier": "artifact_definition___group",
                    "cardinality": "one",
                    "optional": False,
                },
                {
                    "name": "transformation",
                    "peer": InfrahubKind.TRANSFORM,
                    "kind": "Attribute",
                    "identifier": "artifact_definition___transformation",
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "GeneratorDefinition",
            "namespace": "Core",
            "include_in_menu": False,
            "icon": "mdi:state-machine",
            "label": "Generator Definition",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.AWARE.value,
            "uniqueness_constraints": [["name__value"]],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.TASKTARGET],
            "documentation": "/topics/generator",
            "attributes": [
                {"name": "name", "kind": "Text", "unique": True},
                {"name": "description", "kind": "Text", "optional": True},
                {"name": "parameters", "kind": "JSON"},
                {"name": "file_path", "kind": "Text"},
                {"name": "class_name", "kind": "Text"},
                {"name": "convert_query_response", "kind": "Boolean", "optional": True, "default_value": False},
            ],
            "relationships": [
                {
                    "name": "query",
                    "peer": InfrahubKind.GRAPHQLQUERY,
                    "identifier": "generator_definition__graphql_query",
                    "kind": "Attribute",
                    "cardinality": "one",
                    "optional": False,
                },
                {
                    "name": "repository",
                    "peer": InfrahubKind.GENERICREPOSITORY,
                    "kind": "Attribute",
                    "cardinality": "one",
                    "identifier": "generator_definition__repository",
                    "optional": False,
                },
                {
                    "name": "targets",
                    "peer": InfrahubKind.GENERICGROUP,
                    "kind": "Attribute",
                    "identifier": "generator_definition___group",
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "GeneratorInstance",
            "namespace": "Core",
            "label": "Generator Instance",
            "include_in_menu": False,
            "icon": "mdi:file-document-outline",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "branch": BranchSupportType.LOCAL.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.TASKTARGET],
            "documentation": "/topics/generator",
            "attributes": [
                {"name": "name", "kind": "Text"},
                {
                    "name": "status",
                    "kind": "Text",
                    "enum": GeneratorInstanceStatus.available_types(),
                },
            ],
            "relationships": [
                {
                    "name": "object",
                    "peer": InfrahubKind.NODE,
                    "kind": "Attribute",
                    "identifier": "generator__node",
                    "cardinality": "one",
                    "optional": False,
                },
                {
                    "name": "definition",
                    "peer": InfrahubKind.GENERATORDEFINITION,
                    "kind": "Attribute",
                    "identifier": "generator__generator_definition",
                    "cardinality": "one",
                    "optional": False,
                },
            ],
        },
        {
            "name": "StandardWebhook",
            "namespace": "Core",
            "description": "A webhook that connects to an external integration",
            "label": "Standard Webhook",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "include_in_menu": False,
            "icon": "mdi:webhook",
            "branch": BranchSupportType.AGNOSTIC.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
            "attributes": [
                {"name": "shared_key", "kind": "Password", "unique": False, "order_weight": 4000},
            ],
        },
        {
            "name": "CustomWebhook",
            "namespace": "Core",
            "description": "A webhook that connects to an external integration",
            "label": "Custom Webhook",
            "default_filter": "name__value",
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "include_in_menu": False,
            "icon": "mdi:cog-outline",
            "branch": BranchSupportType.AGNOSTIC.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.WEBHOOK, InfrahubKind.TASKTARGET],
            "attributes": [],
            "relationships": [
                {
                    "name": "transformation",
                    "peer": InfrahubKind.TRANSFORMPYTHON,
                    "kind": "Attribute",
                    "identifier": "webhook___transformation",
                    "cardinality": "one",
                    "optional": True,
                    "order_weight": 7000,
                },
            ],
        },
        {
            "name": "Namespace",
            "namespace": "Ipam",
            "description": "A namespace that segments IPAM",
            "label": "IPAM Namespace",
            "default_filter": "name__value",
            "human_friendly_id": ["name__value"],
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "include_in_menu": False,
            "icon": "mdi:format-list-group",
            "branch": BranchSupportType.AWARE.value,
            "inherit_from": [InfrahubKind.IPNAMESPACE],
            "attributes": [
                {"name": "default", "kind": "Boolean", "optional": True, "read_only": True, "order_weight": 9000}
            ],
        },
        {
            "name": "IPPrefixPool",
            "namespace": "Core",
            "description": "A pool of IP prefix resources",
            "label": "IP Prefix Pool",
            "include_in_menu": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
            "attributes": [
                {
                    "name": "default_prefix_length",
                    "kind": "Number",
                    "description": "The default prefix length as an integer for prefixes allocated from this pool.",
                    "optional": True,
                    "order_weight": 5000,
                },
                {
                    "name": "default_member_type",
                    "kind": "Text",
                    "enum": ["prefix", "address"],
                    "default_value": "prefix",
                    "optional": True,
                    "order_weight": 3000,
                },
                {
                    "name": "default_prefix_type",
                    "kind": "Text",
                    "optional": True,
                    "order_weight": 4000,
                },
            ],
            "relationships": [
                {
                    "name": "resources",
                    "peer": "BuiltinIPPrefix",
                    "kind": "Attribute",
                    "identifier": "prefixpool__resource",
                    "cardinality": "many",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "optional": False,
                    "order_weight": 6000,
                },
                {
                    "name": "ip_namespace",
                    "peer": "BuiltinIPNamespace",
                    "kind": "Attribute",
                    "identifier": "prefixpool__ipnamespace",
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "optional": False,
                    "order_weight": 7000,
                },
            ],
        },
        {
            "name": "IPAddressPool",
            "namespace": "Core",
            "description": "A pool of IP address resources",
            "label": "IP Address Pool",
            "include_in_menu": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
            "attributes": [
                {
                    "name": "default_address_type",
                    "kind": "Text",
                    "optional": False,
                    "description": "The object type to create when reserving a resource in the pool",
                    "order_weight": 3000,
                },
                {
                    "name": "default_prefix_length",
                    "kind": "Number",
                    "description": "The default prefix length as an integer for addresses allocated from this pool.",
                    "optional": True,
                    "order_weight": 4000,
                },
            ],
            "relationships": [
                {
                    "name": "resources",
                    "peer": "BuiltinIPPrefix",
                    "kind": "Attribute",
                    "identifier": "ipaddresspool__resource",
                    "cardinality": "many",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "optional": False,
                    "order_weight": 5000,
                },
                {
                    "name": "ip_namespace",
                    "peer": "BuiltinIPNamespace",
                    "kind": "Attribute",
                    "identifier": "ipaddresspool__ipnamespace",
                    "cardinality": "one",
                    "branch": BranchSupportType.AGNOSTIC.value,
                    "optional": False,
                    "order_weight": 6000,
                },
            ],
        },
        {
            "name": "NumberPool",
            "namespace": "Core",
            "description": "A pool of number resources",
            "label": "Number Pool",
            "include_in_menu": False,
            "branch": BranchSupportType.AGNOSTIC.value,
            "generate_profile": False,
            "inherit_from": [InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
            "attributes": [
                {
                    "name": "node",
                    "kind": "Text",
                    "optional": False,
                    "description": "The model of the object that requires integers to be allocated",
                    "order_weight": 3000,
                },
                {
                    "name": "node_attribute",
                    "kind": "Text",
                    "description": "The attribute of the selected model",
                    "optional": False,
                    "order_weight": 4000,
                },
                {
                    "name": "start_range",
                    "kind": "Number",
                    "optional": False,
                    "description": "The start range for the pool",
                    "order_weight": 5000,
                },
                {
                    "name": "end_range",
                    "kind": "Number",
                    "optional": False,
                    "description": "The end range for the pool",
                    "order_weight": 6000,
                },
            ],
        },
        {
            "name": "GlobalPermission",
            "namespace": "Core",
            "description": "A permission that grants global rights to perform actions in Infrahub",
            "label": "Global permission",
            "include_in_menu": False,
            "order_by": ["action__value", "decision__value"],
            "display_labels": ["action__value", "decision__value"],
            "human_friendly_id": ["action__value", "decision__value"],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.BASEPERMISSION],
            "branch": BranchSupportType.AGNOSTIC.value,
            "attributes": [
                {
                    "name": "action",
                    "kind": "Dropdown",
                    "choices": [{"name": permission.value} for permission in GlobalPermissions],
                    "order_weight": 2000,
                },
                {
                    "name": "decision",
                    "kind": "Number",
                    "enum": PermissionDecision.available_types(),
                    "default_value": PermissionDecision.ALLOW_ALL.value,
                    "order_weight": 3000,
                    "description": "Decide to deny or allow the action at a global level",
                },
            ],
        },
        {
            "name": "ObjectPermission",
            "namespace": "Core",
            "description": "A permission that grants rights to perform actions on objects",
            "label": "Object permission",
            "include_in_menu": False,
            "order_by": ["namespace__value", "name__value", "action__value", "decision__value"],
            "display_labels": ["namespace__value", "name__value", "action__value", "decision__value"],
            "human_friendly_id": ["namespace__value", "name__value", "action__value", "decision__value"],
            "uniqueness_constraints": [["namespace__value", "name__value", "action__value", "decision__value"]],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.BASEPERMISSION],
            "attributes": [
                {"name": "namespace", "kind": "Text", "order_weight": 2000},
                {"name": "name", "kind": "Text", "order_weight": 3000},
                {
                    "name": "action",
                    "kind": "Text",
                    "enum": PermissionAction.available_types(),
                    "default_value": PermissionAction.ANY.value,
                    "order_weight": 4000,
                },
                {
                    "name": "decision",
                    "kind": "Number",
                    "enum": PermissionDecision.available_types(),
                    "default_value": PermissionDecision.ALLOW_ALL.value,
                    "order_weight": 5000,
                    "description": (
                        "Decide to deny or allow the action. If allowed, it can be configured for the default branch, any other branches or all "
                        "branches"
                    ),
                },
            ],
        },
        {
            "name": "AccountRole",
            "namespace": "Core",
            "description": "A role defines a set of permissions to grant to a group of accounts",
            "label": "Account role",
            "icon": "mdi:user-badge",
            "include_in_menu": False,
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "human_friendly_id": ["name__value"],
            "generate_profile": False,
            "attributes": [{"name": "name", "kind": "Text", "unique": True}],
            "relationships": [
                {
                    "name": "groups",
                    "peer": InfrahubKind.ACCOUNTGROUP,
                    "optional": True,
                    "identifier": "role__accountgroups",
                    "cardinality": "many",
                    "kind": "Attribute",
                },
                {
                    "name": "permissions",
                    "peer": InfrahubKind.BASEPERMISSION,
                    "optional": True,
                    "identifier": "role__permissions",
                    "cardinality": "many",
                    "kind": "Attribute",
                },
            ],
        },
        {
            "name": "AccountGroup",
            "namespace": "Core",
            "description": "A group of users to manage common permissions",
            "label": "Account group",
            "icon": "mdi:account-group",
            "include_in_menu": False,
            "order_by": ["name__value"],
            "display_labels": ["name__value"],
            "human_friendly_id": ["name__value"],
            "generate_profile": False,
            "inherit_from": [InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICGROUP],
            "branch": BranchSupportType.AGNOSTIC.value,
            "relationships": [
                {
                    "name": "roles",
                    "peer": InfrahubKind.ACCOUNTROLE,
                    "optional": True,
                    "identifier": "role__accountgroups",
                    "cardinality": "many",
                    "kind": "Attribute",
                }
            ],
        },
    ],
}
